/////////////////////////////////////////////////////////////////////////////////////////////
//		 Bluetooth-I/O für Arduino Intervall-Timer
//
//       17.08.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
// UTF-8!

#include "IntervallTimer.h"

#if BLUESERIAL_IN_SOFTWARE
    BufferedSoftwareSerial blueSerial(blueRX, blueTX);
#endif
    bool    debug;

//
// Set input mode:
//   -1: Upload .hex file to external flash
//    0: Accept command characters, optionally preceeded by an unsigned integer number.
//  > 0: Read floating point numbers (format "+/-1.2e-3") such as 
//	1: PID coefficients, terminated by 'P', 'I', or 'D'
//      2: Temperature offsets for sensor/ambient/dew, terminated by 'S', 'U', or 'T'
//      3: PID target temperature, terminated by 'K'
//      4: Confirm Upload to SPI flash 
//      5: Confirm Send Flash to Bootloader (Reset)
//
void IntervallTimer::EnterInputMode(int mode)
{
  inputMode = mode,
  numDigits = 0;
}

//
// Neue Bluetooth-Eingaben auswerten
// 
// Rückgabe: true wenn da, sonst false.
//
bool IntervallTimer::CheckBlue()
{
  int8_t z;
  while (1) {
    if (!blueSerial.available()) {	// Zeichen über BT da?
      return false;			//  nein
    }
    z = blueSerial.read();		//  ja: abholen
    if (inputMode < 0) {		//   Flash-Upload (-1)?
      #if defined(SPIFLASH_CS)
      int8_t r = bootFlash.Add(z);	//    weitergeben,
      if (r > 0) {			//  wenn fertig Reboot:
	EnterNewLineLn(PSTR("Sketch geladen und geprüft: Mit '2F' als Update installieren."));
      }
      else if (r < 0) {			//  Fehler
	EnterNewLine(PSTR("Sketch fehlerhaft geladen: Err "));
	bluePrintln(r);
      }
      else {				//  nicht fertig
	continue;			//  gleich nächstes Zeichen prüfen
      }
      #endif
      EnterInputMode(0);		// Upload beendet
      return false;
    }
    else {
      break;				// normale Eingabe
    }
  }
  if (z == 10  ||  z == 13) {		//   cr, lf komplett ignorieren,
    sentNewline++;			//    aber protokollieren, ob jede Eingabe eines 
    return false;			//    mitbekommt (wenn entsprechend eingestellt)
  }
  sentNewline = 0;
  z = toupper(z);
  if (inputMode != 0) {			// Zahleneingabe für PID, T-Korr. etc.?
    // numDigits = Anzahl Zeichen
    // erstes Z. darf + oder - sein
    // alles außer Digit beendet den Mode

    if ((isDigit(z) || z=='-' || z=='+' || z=='.' || z=='E')) { // Ziffer etc.?
      numberBuffer[numDigits] = z;
      numDigits++;
      if (numDigits < ELEMENTS(numberBuffer)) {	      //  und noch mehr Platz
	return true;
      }
    }
    else {
      numberBuffer[numDigits] = 0;
      float f;
      char* p;
      f = strtod(numberBuffer, &p);	  // (spart ggü. sscanf() knapp 1 KB!!!)
      if (numDigits > 0 && *p == 0) {	  // gültige Zahl?
	#ifdef COOLER_PWM
	while (inputMode == 1) {		  // PID-Koeff?
	  if (z == 'P') {
	    KpKiKd[0] = f;
	  }
	  else if (z == 'I') {
	    KpKiKd[1] = f;
	  }
	  else if (z == 'D') {
	    KpKiKd[2] = f;
	  }
	  else {
	    break;
	  }
	  EnterNewLine(); PrintPID(KpKiKd); PrintNewLine();
	  EnterInputMode(0);
	  return true;
	}
	if (inputMode == 3 && z == 'K') {		  // Ziel-Temp.?
	  EnterNewLine();
	  StartPIDT(f * 10);
	  EnterInputMode(0);
	  return true;
	}
	#endif
	#if defined(T_SENSOR_AVAILABLE) || defined(DEW_AVAILABLE)
	while (inputMode == 2) {
	  #if defined(T_SENSOR_AVAILABLE)
	  if (z == 'S') {		  // Temp.-Korr?
	    tempCorr = int(f * 10);
	  }
	  else
	    #endif
	    #if defined(DEW_AVAILABLE)
	    if (z == 'U') {		  // Tu-Korr?
	      TuCorr = int(f * 10);
	    }
	    else if (z == 'T') {		  // TP-Korr?
	      TdCorr = int(f * 10);
	    }
	    else
	      #endif
	      break;
	  EnterNewLineLn(PSTR("Korrektur übernommen, kann mit 'E' dauerhaft gespeichert werden."));
	  EnterInputMode(0);
	  return true;
	}
	#endif
	#if defined(SPIFLASH_CS)
	if (inputMode == 4  &&  f == 1) {  //  UploadFileToFlash bestätigt?
	  EnterNewLine(PSTR("Lösche Flash (32KB)"));
	  if (bootFlash.Reset()) {
	    bluePrint_P(PSTR(". Bereit: Starten und (grüne) Meldung\n abwarten, vorher keine Eingaben!"));
	    EnterInputMode(-1);
	  }
	  else {
	    bluePrint_P(PSTR("Fehler, Upload abgebrochen!"));
	    EnterInputMode(0);
	  }
	  return false;
	}
	if (inputMode == 5  &&  f == 2) {  // SendFlashToBootloader bestätigt?
	  SendFlashToBootloader(); 
	  return false;
	}
	#endif
        f = f;
      }
    }
    SendErrorBlue();
    EnterInputMode(0);
    return true;
  }
  if (z >= '0'  &&  z <= '9') {		// Zifferneingabe?
    if (HasNumber()) {
      numberInput = numberInput * 10 + (z - '0');
    }
    else {
      numberInput = (z - '0');
    }
    numDigits++;
    return true;
  }
#if defined(COOLER_PWM) || defined(COOLER_POWER)
# define NUMCMNDS "SMLNP%REAFCK"
#else
# define NUMCMNDS "SMLNP%REAF"
#endif
  if (strchr(NUMCMNDS, z) != 0) {	// Befehle mit Zahlenangabe:
    //
    // Zuerst prüfen, ob ein Zeichen, das eine Zifferneingabe übernimmt,
    //  vorliegt und ggf. auswerten:
    //
    if (z == 'S') {    // Sekunden
      if (!HasNumber()) {
	ShowHelpOn_s();
      }
      else if (numberInput == 0) {
	speedIndex = 0;	    // schnelle Auslösung
	SendInfoBlue(1);
      }
      else if (-numberInput < 0) {  // 1..-minInt
	speedIndex = -numberInput;
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == 'M') {    // Minuten
      if (!HasNumber()) {
	ShowHelpOn_m();
      }
      else if (numberInput == 0) {
	speedIndex = numberInput;	    // schnelle Auslösung
	SendInfoBlue(1);
      }
      else if (numberInput <= (0x7FFF / 60)) {
	speedIndex = -(numberInput * 60);
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == 'L') {    // Limit in Minuten
      if (!HasNumber()) {
	ShowHelpOn_l();
      }
      else SetLimit(numberInput);
    }
    else if (z == 'N') {    // Limit = Anzahl Frames
      if (!HasNumber()) {
	ShowHelpOn_n();
      }
      else if (numberInput == 0) {
	SetLimit(0);	  // endlos
      }
      else {
	SetLimit((GetSpeed() * numberInput + 59999L) / 60000L);	// Gesamtzeit (aufgerundet) in msec --> min
      }
    }
    else if (z == 'P') {    // Pause in Zehnteln
      if (!HasNumber()) {
	ShowHelpOn_p();
      }
      else if (numberInput > 0) {
	bulbBreak_ms = numberInput * 100L;  // --> millis
	//
	// min. Bulb-Pause einhalten: bei zu kleinen Pausen startet
	//  nur jede zweite Auslösung!
	/* Bei Bedarf durch Reset einstellen:
	if (bulbBreak_ms < bulbBreak_min) {   // Minimalwert
	  bulbBreak_ms = bulbBreak_min;
	}
	*/
      }
      else {				    // für 0:
        bulbBreak_ms = bulbBreak_min;	    //  kleinsten guten Wert einstellen
      }
      SendInfoBlue(1);
    }
    else if (z == '%') {    // Belichtungszeit variieren, in %
      if (!HasNumber()) {
	ShowHelpOn_pcnt();
      }
      else if (numberInput < 100) {
	speedDither = numberInput;
	speedOffset = 0;
	SendInfoBlue(1);
      }
      else {
	SendErrorBlue();
      }
    }
    else if (z == 'E') {    // Einstellungen
      if (!HasNumber()) {
	EnterNewLine(); StartHelp(settingsHelpPage); 
      }
      else switch (numberInput) {
	case 0:			// aktuelle Settings anzeigen
	  ShowSettings();
	break;
	case 1:
	  ResetSettings(false);
	  EnterNewLine(PSTR("Einstellungen auf Basiswerte zurückgesetzt."));
	  ShowSettings();
	break; 
	case 2:
	  if (ResetSettings(true)) {
	    EnterNewLine(PSTR("Einstellungen aus EEPROM geladen."));
	    ShowSettings();
	  }
	  else {
	    EnterNewLine(PSTR("Es sind keine Einstellungen im EEPROM gespeichert."));
	  }
	break;
	case 3:			// Settings in EEPROM speichern
	  SaveSettingstoEEPROM();
	  EnterNewLine(PSTR("Einstellungen im EEPROM gespeichert."));
	break; 
	case 4:			// Settings aus EEPROM löschen
	  ClearSettingsfromEEPROM();
	  EnterNewLine(PSTR("Einstellungen aus EEPROM gelöscht."));
          SendInfoBlue(1);
	break; 
	default: SendErrorBlue();
      }
    }
#if defined(SPIFLASH_CS)	// 'F'
    else if (z == 'F') {    // Update
      EnterNewLine(); 
      if (!HasNumber()) {
        StartHelp(uploadHelpPage); 
      }
      else switch (numberInput) {
	case 0:	PrintFlash(); // 0F
	break;
	case 4:	PrintFlash(1); // 0F
	break;
	case 1:	if (shootState == stIdle) {
		  bluePrint_P(PSTR("Upload (von ShutterEmuEOS.ino.hex): Bestätigen mit '1F': "));
		  EnterInputMode(4);
		} 
		else {
		  bluePrint_P(PSTR("Uploads sind nur bei inaktiver Kamera möglich!"));
		}
	break;
	case 2:	if (!bootFlash.HasValidUpload()) {
		  bluePrint_P(PSTR("Zuerst Upload des Sketches durchführen!"));
		}
		else {
		  bluePrint_P(PSTR("Upload als Update installieren: Bestätigen mit '2F': "));
		  EnterInputMode(5);
		}
	break;
	case 3: if (bootFlash.Reset()) {  //3F
		  bluePrint_P(PSTR("Update-Flash (32KB) gelöscht."));
		}
		else {
		  bluePrint_P(PSTR("Fehler beim Löschen!"));
		}
	break;
	default: SendErrorBlue();
      }
    }
#endif
  if (z == 'R') {    // Reset
    EnterNewLine(); 
    if (!HasNumber()) {
      StartHelp(resetHelpPage); 
    }
    else switch (numberInput) {
      case 1:	Reset();	    // 1R
		bluePrintln_P(PSTR("Einstellungen wiederhergestellt."));
      break;
      case 2:  watchdogReset();    // 2R
		//bluePrintln_P(PSTR("Bumm!"));
      break;
      default: SendErrorBlue();
    }
  }
#if defined(PROTOCOL)	// 'A'
    else if (z == 'A') {    // Protokoll
      EnterNewLine(); 
      if (!HasNumber()) {
	if (protocol.IsSending()) {
	  if (protocol.IsSendingPaused()) {
	    protocol.ContinueSending();
	  }
	  else {
	    // 
	    // Eine Eingabe bei lfd. Protokollausgabe wird meist
	    // nicht korrekt erkannt! Deshalb nicht hier prüfen,
	    // sondern unten für alle Zeichen!
	    //
	    //protocol.PauseSending();
  	    //EnterNewLineLn(PSTR("Weiter mit 'A', Abbruch mit '2A'"));
	  }
	}
	else {
  	  StartHelp(protocolHelpPage); 
	}
      }
      else switch (numberInput) {
	case 0:	StopProtocol();  break;      // 0A : Protokoll anhalten
	case 1:	ContinueProtocol(); break;   // 1A : Protokoll starten/fortsetzen
	case 2:	RestartProtocol(); break;    // 2A : Protokoll löschen
	case 3:	SendProtocol();  break;      // 3A : Protokoll anzeigen
	default: SendErrorBlue();
      }
    }
#endif
#if defined(COOLER_PWM) || defined(COOLER_POWER)
    else if (z == 'K') {    // Kühlungsoptionen: 0=aus 1=min, 2=low, 3=std, 4=hi, 5=max, 1xxx=xxx, 8=geregelt, 9=geregelt nach 10 min
      if (!HasNumber()) {
	EnterNewLine(); 
	#ifdef COOLER_PWM
	StartHelp(calibHelpPage); 
	#else
	ShowHelpOn_kh();
	#endif
      }
      else switch (numberInput) {
	case 0:	
	case 1:	
	case 2:	
	case 3:	
	case 4:	
	case 5:	StopPID(); SelectCoolingMode(numberInput); SendInfoBlue(1); break;
	case 7:	bluePrint_P(PSTR(" Ziel-Temperatur eingeben, mit 'K' abschließen: "));
		EnterInputMode(3);
		break;
	case 8:	StartPID(0);   EnterNewLine(); break;
	case 9:	StartPID(600); EnterNewLine(PSTR("Temperaturregelung startet in 10 Minuten.")); 
		break;
	#if  defined(DEW_AVAILABLE)
	case 10: 
	case 11: EnableDewControl(numberInput==11); EnterNewLine(PSTR("Taupunktkontrolle ")); 
		if (IsDewControl()) {
		  bluePrintln_P(PSTR("aktiv"));
		}
		else {
		  bluePrintln_P(PSTR("aus"));
		}
		break;
	#endif
	#if defined(DEW_AVAILABLE) && (defined(HEATER_PWM) || defined(HEATER_SOFT_PWM))
	case 29: nextDewCheck = uint8_t(millis() >> 10);	// Heizung an Taupunkt anpassen: Start jetzt (Zeit in Sekunden)
		[[fallthrough]];
	#endif 
	#if defined(HEATER_POWER)
	case 21: 
	# if defined(HEATER_PWM) || defined(HEATER_SOFT_PWM)
	case 22: 
	case 23: 
	case 24: 
	case 25: 
	# endif
	case 20: SelectHeatingMode(numberInput-20); SendInfoBlue(1);
	break;
	break;
	#endif
	default: 
	  if (numberInput >= 1000  &&  numberInput <= 1255) {
	    StopPID(); 
	    StartCooling((uint8_t)(numberInput - 1000));
	    SendInfoBlue(1);
	  }
	  else {
  	    SendErrorBlue(); 
	  }
	  break;
      }
    }
#endif
#ifdef COOLER_PWM
    else if (z == 'C') {    // Kalibrierung
      if (!HasNumber()) {
	#if defined(CALIB_PID)
	if (IsCalibPID()) {    // "PID-Kalibirerung aus" geht auch ohne
	  UpdatePIDCalib(true);	  // stop
	}
	else 
        #endif
	{
	  EnterNewLine(); StartHelp(calibHelpPage); 
	}
      }
      else switch (numberInput) {
	case 0:	EnterNewLine(); PrintPID(KpKiKd); PrintNewLine(); break; // Print PID
	case 1:			        // PID auf Basiseinstellung
	  ResetPID(false);
	  EnterNewLine(PSTR("PID-Koeffizienten auf Basiswerte zurückgesetzt."));
	break;
	case 2:				// Reset PID
	  if (ResetPID(true)) {
	    EnterNewLine(PSTR("PID-Koeffizienten aus EEPROM geladen."));
	  }
	  else {
	    EnterNewLine(PSTR("Es sind keine PID-Koeffizienten im EEPROM gespeichert."));
	  }
	break;
	case 3:			// Settings in EEPROM speichern
	  SavePIDtoEEPROM();
	  EnterNewLine(PSTR("PID-Koeffizienten im EEPROM gespeichert."));
	break; 
	case 4:				// PID auf Basiseinstellung, aus EEPROM löschen
	  ClearPIDfromEEPROM();
	  EnterNewLine(PSTR("PID-Koeffizienten im EEPROM gelöscht."));
	break; 
	case 5:		    	// Enter PID
	  bluePrint_P(PSTR(" PID-Koeffizient eingeben, mit P/I/D abschließen: "));
	  EnterInputMode(1);
	break; 
	default:		        // Start PID Calib
        #if defined(CALIB_PID)
	  EnterNewLine(); 
	  CalibPID(numberInput);
	#else
  	  SendErrorBlue(); 
	#endif
	break; 
      }
    }
#endif
  }
#if defined(PROTOCOL)
  else if (protocol.IsSending()) {	
    // 
    // Eine Eingabe bei lfd. Protokollausgabe wird meist
    // nicht korrekt erkannt (s.o. bei 'A'):
    // Deshalb hier für alle Zeichen unterbrechen:
    // 
    protocol.PauseSending();
    EnterNewLineLn(PSTR("Weiter mit A, Abbruch mit 3A"));
  }
#endif
  else if (HasNumber()) {			// die weiteren Kommandos sind nur erlaubt, wenn keine Zifferneingabe vorausgeht
    EnterNewLine(PSTR("Zahleneingabe hier nicht erlaubt!"));
  }
#if defined(T_SENSOR_AVAILABLE) || defined(DEW_AVAILABLE)
  else if (z == 'T') {
    bluePrint_P(PSTR(" Temperatur-Korrektur eingeben, abschließen mit "));
#if defined(T_SENSOR_AVAILABLE)
    bluePrint_P(PSTR("'S' für Sensor, "));
#endif
#if defined(DEW_AVAILABLE)
    bluePrint_P(PSTR("'U' für Umgebung, "));
    bluePrint_P(PSTR("'T' für Taupunkt, "));
#endif
    bluePrint_P(PSTR(": "));
    EnterInputMode(2);
  }
#endif
  else if (z == '!'  ||  z == '#') {
    if (shootState == stIdle) {	        // aus?
      shootOnceHappend = 0;		// Shutter hat noch nix gemeldet
      if (z == '!') {			// neue Auslösung:
	speedOffset = 0;		// Dither-Start
	StartFrame();			//  ab jetzt Shoot wie gewählt wiederholen
	SetShootLimit();		// Limit-Zeit eintragen
      }
      else {				// '#'
  	shootState = stShootBulb;	
	SetShootLine(true);		//   los!
	endShoot = millis();		//   ab jetzt
	nextUpdate = endShoot + 1000;
      }
    }
    else {				// schon an: abschalten
      //
      // '#' schaltet sofort ab, '!' nur bei Dauer- oder Schnell-Auslösung.
      //
      if (shootState == stShootBulb) {	// Kamera auf Bulb?
        SetShootLine(false);		//  sofort abschalten
	if (quiet) {			//  Zeit anzeigen: nur wenn quiet nötig
          bluePrint_P(PSTR(" ("));
	  blueSerial.print((int)((millis() - endShoot)/1000));
          bluePrintln_P(PSTR("s)"));
	}
	shootState = stIdle;
      }
      else if (IsFastSpeed()) {	    	// schnelle Auslösung einfach nur nicht wiederholen
	shootState = stIdle;
      }
      else {
	if (z == '!') {
	  if (RequestSeriesTermination()) {	// !: fertig, wenn Auslösung regulär beendet
	    bluePrint_P(PSTR("- wird beendet -"));
	  }
	  else {
	    bluePrint_P(PSTR("- wird fortgesetzt -"));
	  }
	}
	else {				// #: sofort abschalten
	  SetShootLine(false);
    	  shootState = stIdle;
	  int32_t dauer = GetSpeed() - (endShoot - millis());
	  if (dauer > 0) {		// negativ, wenn t verkürzt wurde
	    bluePrint_P(PSTR(" ("));
	    blueSerial.print((int)(dauer / 1000));
  	    bluePrintln_P(PSTR("s)"));
	  }
	}
      }
      if (shootState == stIdle) {
	EnterNewLine();
      }
    }
  }
  else if (z == '+'  ||  z == '-') {	// Belichtungsschritt:
    if (speedIndex < 0) {		// bisher direkt in Sekunden angegeben?
      uint32_t spd_ms = GetSpeed();	//  in msec in der Tabelle suchen
      if (z == '+') {			//   von unten her, wenn '+':
	speedIndex = 1;
	do {
	  if (GetSpeed() > spd_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert überschritten hat
	  }
	  speedIndex++;
	} while (speedIndex < GetNumSpeeds());
      }
      else {	      			//   von oben her, wenn '-':
	speedIndex = GetNumSpeeds() - 1;
	do {
	  if (GetSpeed() < spd_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert unterschritten hat
	  }
	  speedIndex--;
	} while (speedIndex > 1);
      }
    }
    else {				// Belichtung nach Tabellenwert: um 1 Schritt ändern
      if (z == '+') {
	if (speedIndex < GetNumSpeeds() - 1) {	  //  noch nicht durch?
	  speedIndex++;
	}
      }
      else {
	if (speedIndex > 0) {
	  speedIndex--;
	}
      }
    }
    SendInfoBlue(1);
  }
  else if (z == '>'  ||  z == '<') {	// Zeitbegrenzung
    if (limitIndex < 0) {		// bisher direkt in Minuten angegeben?
      uint32_t lim_ms = GetLimit();	//  bisheriges Limit in msec
      // 
      // Passenden Wert in der Tabelle suchen:
      // Zur Verkleinerung ('<') von oben her (d.h. in der Tabelle von 
      //  links nach rechts) suchen, bis der Tabellenwert kleiner als der
      //  bisherige Wert ist.
      // Zur Vergrößerung ('>') von unten her (d.h. in der Tabelle von hinten).
      // 
      if (z == '<') {			//  '<': von oben her:
	limitIndex = 1;			//   ([0] = 0 == endlos!)
	do {
	  if (GetLimit() < lim_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert unterschritten hat
	  }
	  limitIndex++;
	} while (limitIndex < GetNumLimits());
      }
      else {	      			//   '>': von unten her:
	limitIndex = GetNumLimits() - 1;//   letzter Wert
	do {
	  if (GetLimit() > lim_ms) {
	    break;			//   fertig, wenn neue Einstellung den alten Wert überschritten hat
	  }
	  limitIndex--;
	} while (limitIndex > 1);	// ([0] = 0 == endlos!)
      }
    }
    else {				// Belichtung nach Tabellenwert: um 1 Schritt ändern
      if (z == '>') {		// erhöhen:
	if (limitIndex > 0) {	//  noch nicht durch?
	  limitIndex--;
	}
      }
      else {
	if (limitIndex < GetNumLimits() - 1) {  //  noch nicht durch?
	  limitIndex++;
	}
      }
    }
    SetShootLimit();		  	  // aktuelle Limit-Zeit updaten (wird bei Auslösung erneuert)
    SendInfoBlue(2);
  }
  else if (z == '?') {    // Info
    SendInfoBlue(-1);
  }
  else if (z == 'U') {    // Updates
    quiet = !quiet;
    if (shootState == stIdle) {
      SendInfoBlue(-1);
    }
    else {
      EnterNewLine();		  // frische Zeile
    }
  }
  else if (z == 'D') {    // Debug
    debug = !debug;
    if (shootState == stIdle) {
      SendInfoBlue(-1);
    }
    else {
      EnterNewLine();		  // frische Zeile
    }
  }
  else if (z == 'H') {    // Hilfe
    StartHelp();
  }
  else {
    SendErrorBlue();
  }
  numDigits = 0;
  return true;
}

////////////////////////////////////////////////////////////////////////////////////
//
// String direkt aus dem PROGMEM an blueSerial ausgeben
//
void bluePrint_P(PGM_P s)
{
  char c;
  while ((c = pgm_read_byte(s)) != 0) {
    blueSerial.write(c); s++;
  }
}

// .. mit newline
inline
void bluePrintln_P(PGM_P s)	   
{
  bluePrint_P(s); 
  bluePrintln();
}


//
// Eine mit 10^dcms skalierte Zahl als Dezimalzahl an blueSerial ausgeben.
// Die Zahl der Dezimalen wird um die Zahl der Vorkommastellen - 1
//  reduziert auf minimal eine.
//
void bluePrintDecimals(int val, int dcms)
{
  if (val < 0) {
    blueSerial.write('-');
    val = -val;
  }
  char bf[30];
  int n = 0;
  while (val != 0) {    // Ziffernfolge von rechts nach links:
    int r = val;
    val = val / 10;
    bf[n++] = (char)(r - val*10 + '0');
  }
  while (n <= dcms) {	// mit 0 auffüllen bis zur ersten Vorkommastelle
    bf[n++] = '0';
  }
  int i = n;
  while (i > 0) {	// Ziffern rückwärts ausgeben:
    if (i == dcms) {	// noch genau dcms Ziffern?
      blueSerial.write('.');
      // n-i - 1 = Zahl der angez. Vorkommastellen - 1
      // Von den restlichen i Stellen, den Dezimalen, abziehen
      //  --> i - (n-i-1) = 2i - n + 1
      //  aber mindestens eine:
      //
      i = max(1, 2*i - n + 1);
    }
    i--;
    blueSerial.write(bf[i]);
  }
}

//
// Temperatur ausgeben
//
void bluePrintTdegC(int Tx10)
{
  bluePrintDecimals(Tx10, 1);
  bluePrint_P(PSTR("°C"));
}

//
// Temperatur mit Vorzeichen ausgeben
//
void bluePrintTdegCsign(int Tx10)
{
  if (Tx10 >= 0) {
    bluePrint_P(PSTR("+"));
  }
  bluePrintTdegC(Tx10);
}

////////////////////////////////////////////////////////////////////////////////////
#if BLUESERIAL_IN_SOFTWARE
//
// Zahl in den Puffer schreiben
//
void BufferedSoftwareSerial::print(long l)
{
  if (l < 0) {
    write('-');
    l = -l;
  }
  print((unsigned long)l);
}

void BufferedSoftwareSerial::print(unsigned long l)
{
  if (l == 0) {
    write('0');
    return;
  }
  char bf[30];
  int x = 0;
  while (l != 0) {    // Ziffernfolge von rechts nach links:
    long r = l;
    l = l / 10;
    bf[x++] = (char)(r - l*10 + '0');
  }
  while (x > 0) {     // Ziffern rückwärts ausgeben:
    x--;
    write(bf[x]);
  }
}

void BufferedSoftwareSerial::print(int l)
{
  if (l < 0) {
    write('-');
    l = -l;
  }
  print((unsigned int)l);
}

void BufferedSoftwareSerial::print(unsigned int l)
{
  if (l == 0) {
    write('0');
    return;
  }
  char bf[30];
  int x = 0;
  while (l != 0) {    // Ziffernfolge von rechts nach links:
    int r = l;
    l = l / 10;
    bf[x++] = (char)(r - l*10 + '0');
  }
  while (x > 0) {     // Ziffern rückwärts ausgeben:
    x--;
    write(bf[x]);
  }
}

void BufferedSoftwareSerial::printHex(uint16_t v)
{
  for (int i=sizeof(v)*8; i>0; ) {	// 16 Bit
    i -= 4;
    int8_t d = (v >> i) & 0xf;
    if (d > 9) {
      d += 'A' - '0' - 10;
    }
    write(d + '0');
  }
}

void BufferedSoftwareSerial::printHex(uint8_t v)
{
  for (int i=sizeof(v)*8; i>0; ) {	// 8 Bit
    i -= 4;
    int8_t d = (v >> i) & 0xf;
    if (d > 9) {
      d += 'A' - '0' - 10;
    }
    write(d + '0');
  }
}

#endif //BLUESERIAL_IN_SOFTWARE

