#ifndef BufferedSoftwareSerial_h_
#define BufferedSoftwareSerial_h_

// <-- Comment out to force SoftwareSerial buffer into internal RAM
#if defined(SPIFLASH_CS)		// Flash verfügbar:
#define BSS_FLASH
#endif
//*/

#if defined(BSS_FLASH)
#include "FlashUpdate.h"
# define BSS_T  uint16_t
# define BSS_BUFFERSIZE  0x8000U	// 2 Sektoren a 16K
#else
# define BSS_T  uint8_t
# define BSS_BUFFERSIZE  (BSS_T(200U))
#endif

// Anbindung des Bluetooth-Moduls (HC-05) über Software-Serial() mit Sende-Puffer:
//
#include <SoftwareSerial.h>
//
// SoftwareSerial hat keine TX-Puffer. Dadurch blockieren längere Sendungen
//  den gesamten Programmablauf.
// BufferedSoftwareSerial implementiert einen TX-Puffer, in den die benötigten
//  print() Funktionen ihre Ausgabe schreiben. Das Senden muss dann in loop()
//  Zeichen für Zeichen mit SendNextChar() veranlasst werden. Je nach Dauer
//  eines loop() Durchlaufs sinkt die effektive Baudrate dann massiv ab.
//
class BufferedSoftwareSerial : protected SoftwareSerial {
  public:
    BufferedSoftwareSerial(int rx, int tx) : SoftwareSerial(rx, tx)
    {
      iWrite = 0;
      iRead  = 0;
    }

    bool    SendNextChar();
    void    println()		  {write(10);}
    void    print(const char* s);
    void    print(int i);
    void    print(unsigned int i);
    void    print(long l);
    void    print(unsigned long l);
    void    printHex(uint8_t v);
    void    printHex(uint16_t v);
    void    println(const char* s)  {print(s); println();}
    void    println(int i)	    {print(i); println();}
    void    println(unsigned int i) {print(i); println();}
    void    println(long l)	    {print(l); println();}
    void    println(unsigned long l){print(l); println();}
    bool    write(char c);
    void    writeThrough(char c)    {SoftwareSerial::write(c);}

    BSS_T   GetFreeCount() const;
  static BSS_T GetHalfSize()        {return (BSS_T)(BSS_BUFFERSIZE / 2);}

    // von außen zugängliche Funktionen der Basisklasse:
    //
    using SoftwareSerial::read;
    using SoftwareSerial::available;
    using SoftwareSerial::begin;

  protected:
    void    store(char c);
    char    fetch();
    bool    IsBufferEmpty() const   {return iWrite == iRead;}

  protected:
    BSS_T   iWrite;
    BSS_T   iRead;
    #if !defined(BSS_FLASH)
    char    buffer[BSS_BUFFERSIZE];
    #endif
};

//
// Zahl der freien Plätze im Puffer bestimmen.
//
inline
BSS_T BufferedSoftwareSerial::GetFreeCount() const     
{
  /*
  int charcnt = iWrite - iRead;		  // Pufferinhalt
  if (charcnt >= 0) {			  // wenn die Schreibpos. vor der Lesepos. ist
    return BSS_BUFFERSIZE - charcnt;	  //  
  }
  return charcnt;	  // sonst ist charcnt der Verbrauch
  */
  if (iWrite >= iRead) {	  
    // die Schreibpos. hat das Ende noch nicht erreicht 
    //  und ist vor der Lesepos.:
    //
    return BSS_T(BSS_BUFFERSIZE - (iWrite - iRead)); 
  }
  // die Schreibpos. hat das Ende überschritten 
  //  und die Lesepos. noch nicht.
  // --> das Stück dazwischen ist frei
  return iRead - iWrite;
}

//
// Nächstes gepuffertes Zeichen senden.
// 
// Rückgabe: false wenn Puffer leer war.
//
inline
bool BufferedSoftwareSerial::SendNextChar()
{
  if (IsBufferEmpty()) {
    return false;
  }
  SoftwareSerial::write(fetch());	  // nächstes Zeichen mit write() der Basisklasse abschicken
  iRead++;
  if (iRead >= BSS_BUFFERSIZE) {
    iRead = 0;
  }
  return true;	
}


//
// Zeichen in den Puffer schreiben. 
//
inline
void BufferedSoftwareSerial::store(char c)
{
  #if defined(BSS_FLASH)	// Flash verfügbar:
  if ((iWrite & 0x0FFF) == 0) {	// neuer 4K-Block?
    bootFlash.eraseSector(SERIALOUT_FLASHPAGE | iWrite);  // sector erase time is 400ms
  }
  bootFlash.writeChar(SERIALOUT_FLASHPAGE | iWrite, c);
  #else
  buffer[iWrite] = c;
  #endif
}


//
// Zeichen aus dem Puffer holen
//
inline
char BufferedSoftwareSerial::fetch()
{
  return
  #if defined(BSS_FLASH)	// Flash verfügbar:
  bootFlash.readChar(SERIALOUT_FLASHPAGE | iRead);
  #else
  buffer[iRead];
  #endif
}


//
// Zeichen in den Puffer schreiben. Überschreibt/ersetzt SoftwareSerial::write(int8_t i)
//
// Rückgabe: false wenn der Puffer überläuft.
//
inline
bool BufferedSoftwareSerial::write(char c)
{
  store(c);
  iWrite++;
  if (iWrite >= BSS_BUFFERSIZE) {
    iWrite = 0;
  }
  if (iWrite == iRead) {
    if (iWrite == 0) {
      iWrite = BSS_BUFFERSIZE;
    }
    iWrite--;
    return false;
  }
  return true;
}


//
// String in den Puffer schreiben
//
inline
void BufferedSoftwareSerial::print(const char* s)
{
  char c;
  while ((c = *s) != 0) {write(c); s++;}
}


#endif
