/////////////////////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.05.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////

#include "IntervallTimer.h"

#if defined(SPIFLASH_CS)	    // indicates SPI flash module available

#include "FlashUpdate.h"
 
BootFlash bootFlash(SPIFLASH_CS);

//#define DEBUG_FLASH	  // Achtung: Fördert dynamic memory shortage!

// Zugriff auf den Flashspeicher init.
void IntervallTimer::setupSPIFlash()
{
  bootFlash.begin();
  #if defined(DEBUG_FLASH)
  Serial.begin(57600);
  #endif
}


//
// Upload in Flash initialisieren.
// Die ersten 32K des Flash löschen, wenn eraseFlash==true (default).
// 
// Rückgabe: False wenn Fehler beim Löschen
//
bool BootFlash::Reset(bool eraseFlash)		      
{
  digitCnt = -1; 
  if (eraseFlash) {
    uploadState = 0;
    if (!eraseBlock32K(0)) {
      for (int i = 0; i < 8; i++) {  // sector erase time is 400ms
	if (!eraseSector(i << 12)) {	    // 4K
	  return false;
	}
      }
    }
  }
  return true;
}


void IntervallTimer::PrintFlash(bool pmem)
{
  for (uint16_t x = 0 ; x < 0x0100 ; x++) {
    if ((x & 0xf) == 0) {
      bluePrintln();
      bluePrintHex(x);
      bluePrint_P(PSTR(": "));
    }
    if (pmem) {
      bluePrintHex(pgm_read_byte(x));
    }
    else {
      bluePrintHex(bootFlash.readByte(x));
    }
    bluePrint_P(PSTR(" "));
  }
  blueSerial.println();
}


bool IntervallTimer::SendFlashToBootloader()
{
  if (!bootFlash.WriteFirstBytes()) {
    EnterNewLine(PSTR("Fehler beim Flash-Zugriff!"));
    return false;
  }
  //EnterNewLine(PSTR("Reset wird ausgeführt .."));
  watchdogReset();
  return true;
}

// Neues Zeichen aus der .hex Datei aufnehmen und ggf. verarbeiten.
// (s. https://en.wikipedia.org/wiki/Intel_HEX)
// 
// Rückgabe: 1, wenn fertig, 0 wenn weitere Daten erwartet, < 0 bei Fehlern:
//  -1 : Record zu groß
//  -2 : Checksum
//  -3 : Record Typ unbekannt
//  -4 : Flash Write Error
//  -5 : Ungültiges Zeichen
//
int8_t BootFlash::Add(int8_t c)
{
  if (digitCnt < 0) {
    if (c == ':') {	      // neuer Record
      digitCnt = 0;
      return 0;
    }
    if (isspace(c)) {	      // cr, lf vor Records ignorieren
      return 0;
    }
    return -5;		      // andere Zeichen vor Records nicht erlauben
  }
  const char* hexdigs = "0123456789ABCDEF";
  uint8_t val = (uint8_t)(strchr(hexdigs, c) - hexdigs);
  if (val < 16) {	      // Hex-Ziffer?
    uint8_t* p = &lineBuffer[digitCnt >> 1];
    digitCnt++;		      // gelesene Digits (= 2 x Bytes)
    if (digitCnt & 1) {	      // ungerade?
      *p = val << 4;	      //  1. Digit eintragen
      if (digitCnt == 1) {    //  byte count noch nicht da?
	return 0;	      //   dann noch nichts zu tun
      }
    }
    else {		      // gerade Zahl von Digits:
      val += *p;	      //  Digits zu Byte zusammenfassen
      *p = val;
      if (digitCnt == 2) {    //  byte count gerade erhalten? (in lineBuffer[0])
	limit = (1 + 2 + 1 + GetByteCount() + 1) << 1;	// Byte count + Address + Record type + Data + Checksum
	if (limit > (ELEMENTS(lineBuffer) << 1)) {
	  return -1;	      // Record zu groß!
	}
	checksum = val;
	/*  führt zu Fehler beim Upload weil zu langsam:   !!!!!!!!!!
	const int pcnt = 1024 / 16;   // 16-Byte Records pro KB
	static int np = pcnt;
	if (np == 0) {
	  //blueSerial.write('.');      // dauert zu lange
	  blueSerial.writeThrough('.'); // auch nicht besser
	  np = pcnt;
	}
	np--;
	//*/
	return 0;
      }
      checksum += val;	      // Summe über alles (incl. Checksum)
    }
    if (digitCnt == limit) {  // Record vollständig:
      if (checksum != 0) {    // (Checksum ist 1 - byte_summe !)
	return -2;
      }
      if (GetRecordType() == 1) {
	uploadState = 1;      // zeigt an, dass Upload erfolgreich war
	return 1;	      // fertig (falls die Datei noch weiter geht, wird sie normaler Input!)
      }
      if (GetRecordType() != 4) { // 32-Bit Adresse: ignorieren, wird zur Kennzeichnung von ...ino.with_bootloader.hex verwendet.
	if (GetRecordType() != 0) { // (s. https://en.wikipedia.org/wiki/Intel_HEX)
	  return -3;	      // sollte nicht vorkommen
	}
	uint8_t* pd = GetData();
	uint8_t size = GetByteCount();
	uint16_t addr = GetAddress();
	// Write to flash:
	if (addr == 0) {
	  firstBytes[0] = *pd++;     // erstes und zweites Byte noch nicht, sonst wird beim nächsten Reset
	  firstBytes[1] = *pd++;     // das Update durchgeführt, auch wenn etwas schiefgeht
	  size -= 2;
	  addr = 2;
	}
	/* Nur für Vektor-Bootloader:
	else if (addr == 0x0060) {
	  //
	  // Byte 0x0066 patchen: Wird sonst von avrdude oder bootloader erledigt
	  //
	  pd[6] -= 0xFC - 0xD4;	  // 0xFC -> 0xD4, manchmal auch 0xFD -> 0xD5
	}
	//*/
	if (!writeByteArray(addr, pd, size, true)) {
	  return -4;
	}
      }
      digitCnt = -1;	      // nächster Record
    }
    return 0;		      // Record geht noch weiter
  }
  return -5;		      // andere Zeichen sind innerhalb eines Records nicht erlaubt
}

#endif

