/////////////////////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.05.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef IntervallTimer_h_
#define IntervallTimer_h_


#define ELEMENTS(t)	(sizeof(t)/sizeof((t)[0]))

#include <arduino.h>
#include "digitalWriteFast.h"


// Verteilung des externen Flashs:
// 'Page' = 64KB Block
// Page 0 ist für Udates reserviert
//
#define SERIALOUT_FLASHPAGE   0x10000
#define PROTOCOL_FLASHPAGE    0x20000


#if defined(AHT20_AVAILABLE)
# define DEW_AVAILABLE
#endif


// <-- Comment out to skip protocol routines   (frees ~1320 Bytes of program storage space)
#if defined(T_SENSOR_AVAILABLE) || defined(DEW_AVAILABLE)
# define PROTOCOL
# if defined(T_SENSOR_AVAILABLE) && defined(DEW_AVAILABLE)
#  define PROTOCOL_2T	      // include ambient temperature in protocol (380 additional code bytes)
# endif
# include "protocol.h"
#endif
//*/

/*/ <-- Comment out to skip PID calibration routines (frees ~4000 Bytes of program storage space)
#if defined(COOLER_PWM)
#define CALIB_PID
#include "StepResponse.h"
#endif
//*/

#if defined(HEATER_SOFT_PWM) && !defined(HEATER_POWER)
#error Für HEATER_SOFT_PWM wird HEATER_POWER benötigt
#endif


#include "BluetoothIO.h"


#include "cooling.h"
#if defined(CALIB_PWM)
//#include "StepResponse.h"
#endif

#if defined(SPIFLASH_CS)
#include "FlashUpdate.h"
#endif

///////////////////////////////////////////////////////////////

// Faktoren zur Umrechnung von Zeiten in die Basiseinheit Millisekunde
// Vorsicht vor Konflikt mit <ms> in den Shutteremus, das ms in µs umrechnet!
//
#define sec *1000L     		// Zeit von Sekunden in die Basiseinheit (ms) umrechnen
#define Min *60000L    		// Zeit von Minuten in die Basiseinheit (ms) umrechnen
#define h   *3600000L  		// Zeit von Stunden in die Basiseinheit (ms) umrechnen


enum tempModeFlags {
  tmfAuto = 0x01,
  tmfOff  = 0x80
};

///////////////////////////////////////////////////////////////
// Data for IntervallTimer class that can be stored in EEPROM
//  
struct ivtSettings {
    void    SaveSettingstoEEPROM();
    bool    LoadSettingsfromEEPROM();
    void    ClearSettingsfromEEPROM();
  
  protected:
    int	    speedIndex;	      // Timer-Einstellung
    int	    speedDither;      // Timer-Variation
    int	    limitIndex;	      // Wiederholdauer-Einstellung
    int32_t bulbBreak_ms;     // Signal-Unterbrechung zwischen Bulb-Auslösungen
    bool    quiet;
    uint8_t coolingLevel;
    int8_t  coolingMode;
    uint8_t heatingLevel;
    int8_t  heatingMode;
    int	    tempCorr;	      // corrective addition to the sensor temperature
    int	    TuCorr;	      // corrective addition to the ambient temperature
    int	    TdCorr;	      // corrective addition to the dew point temperature
};


class IntervallTimer : public ivtSettings {
  public:
	    IntervallTimer(bool warnOnShootMissing);
	   ~IntervallTimer();

    void    Setup();
    bool    CheckBlue();
    void    Advance();
    void    NotifyShoot();
    bool    SendUpdate();
    int	    UpdateCooling();
    bool    SendNextBlue();

  protected:
    #ifdef COOLER_PWM
    void    StartPID(int afterSeconds);
    void    StartPIDT(float zielX10);
    void    InitPID(int tempX10);
    void    PrintPIDTemp();
    void    StopPID();
    void    AbortPID();
    void    DeletePID();
    bool    IsPIDActive() const;
    bool    IsPIDOff() const;
    #if defined(CALIB_PID)
    bool    CalibPID(int step);
    bool    IsCalibPID() const		    {return PID_delay == -2;}
    int	    UpdatePIDCalib(bool stop);
    void    ContinueCalib();
    #endif
    bool    ResetPID(bool fromEEPROM);
   static void PrintPID(float*);
    void    SavePIDtoEEPROM();
    bool    LoadPIDfromEEPROM();
    void    ClearPIDfromEEPROM();
    void    StartCooling(uint8_t level);
    void    SetCoolingLevel(uint8_t level);
    void    IncreaseCoolingLevelBy(int8_t step);
    #endif
    #if defined(COOLER_PWM) || defined(COOLER_POWER)
    void    SelectCoolingMode(int mode);
    bool    IsCooling() const	    	    {return (coolingMode & tmfOff) == 0;}	  // (gleichw. zu '>= 0') (0 or 255 is lowest level, but not off)
    bool    IsDewControl() const	    {return coolingMode & tmfAuto;}
    bool    IsDewControlActive() const	    {return IsDewControl() && IsCooling();}
    #endif
    #if defined(HEATER_PWM) || defined(HEATER_POWER)
    bool    IsHeating() const	    	    {return (heatingMode & tmfOff) == 0;}	  // heatingLevel 0 means off too
    bool    IsDewHeating() const    	    {return heatingMode & tmfAuto;}
    bool    IsDewHeatingActive() const 	    {return IsDewHeating() && IsHeating();}
    void    SelectHeatingMode(int mode);
    void    SetHeatingLevel(uint8_t level);
    void    IncreaseHeatingLevelBy(int8_t step);
    #endif
    void    WriteHeatingLevel();
    void    WriteCoolingLevel();

    bool    ReadAmbientTemperatureX10(int& temp);
    bool    ReadAmbientHumidity(float& humidity);
    void    TriggerAmbientSensor();
    void    CycleAmbientSensor();
    int	    CalcDewPointX10(const float& humidity, int TuX10);
    void    StopCoolingIdleCam(uint8_t when);
    void    EnableDewControl(bool en);

    void    Reset();
    void    watchdogReset();
    bool    ResetSettings(bool fromEEPROM);
    void    SetShootLimit();
    bool    HasShootLimit() const;
    bool    RequestSeriesTermination();
    bool    CheckShootLimit(uint32_t when);
    void    PrintFrameSummary(int frameOfs);
    void    SetShootLine(bool active);
    void    RandomSpeedOffset();
    void    StartFrame();
    uint32_t GetSpeed() const;
    uint32_t GetLimit() const;
    bool    SetLimit(int32_t minutes);
    int	    GetNumSpeeds() const;
    int	    GetNumLimits() const;
    bool    IsFastSpeed() const	    {return speedIndex == 0;}	      // Timer für schnelle Auslösung wenn nicht Bulb

    #if defined(SPIFLASH_CS)
    bool    SendFlashToBootloader();
    void    setupSPIFlash();
    void    PrintFlash(bool pmem=false);
    #endif

    void    EnterInputMode(int mode);
    bool    HasNumber() const	    {return numDigits != 0;}

    void    SendInfoBlue(int which);
    void    PrintNewLine();
    void    EnterNewLine();
    void    EnterNewLine(PGM_P s);
    void    EnterNewLineLn(PGM_P s);
    void    SendErrorBlue();
    
    void    PrintMore(bool more);
    void    PrintCoolerStatus();
    void    PrintHeaterStatus();
    void    PrintSpeed();
    void    ShowSettings();
    int	    ShowSettingsPart2();
    void    ShowHelpOn_s();
    void    ShowHelpOn_m();
    void    ShowHelpOn_l();
    void    ShowHelpOn_n();
    void    ShowHelpOn_p();
    void    ShowHelpOn_pcnt();
    void    ShowHelpOn_stat();
    #if !defined(COOLER_PWM) && defined(COOLER_POWER)
    void    ShowHelpOn_kh();
    #endif
    void    ShowHelpOn_h();
    int     SendHelp(int what);
    void    StartHelp(int page = 1)  	    {helpPage = page;}
    enum {calibHelpPage = 100, settingsHelpPage = 110, protocolHelpPage = 120, 
	  uploadHelpPage = 130, resetHelpPage = 140, settingsPart = 200};

  private:
    uint32_t bgnShoot;		// Beginn der laufenden Auslösung
    uint32_t endShoot;		// Ende der laufenden Auslösung
    uint32_t shootLimit;	// Ende der laufenden Serie
    int	    shootState;		// Status der Auslösung
    int	    numFrames;
    const int32_t bulbBreak_min = 200;    // min. Bulb-Pause: bei zu kleinen Pausen startet nur jede zweite Auslösung
    const int32_t shootBreak_ms = 100;    // Signal-Unterbrechung zwischen Auslösungen, 2-fach für min. Bulb-Pausen
    int32_t nextUpdate;
    int	    speedOffset;	// aktuelle Timer-Variation
    int	    helpPage;
    unsigned numDigits;
    char    numberBuffer[10];
    int	    numberInput;
    int8_t  inputMode;
    int8_t  sentNewline;
    bool    shootHappend;	// wird gesetzt, wenn Shutter eine Auslösung meldet
    bool    warnShootMissing;	// melden, wenn Auslösungen fehlen: wird ggf. im c'tor gesetzt
    uint8_t shootOnceHappend;	// zeigt an, dass in der aktuellen Sequenz Auslösungen gemeldet wurden
    #if defined(DEW_AVAILABLE)
    uint8_t nextDewCheck;	// Zeitpunkt für nächste Taupunktkontrolle in sec
    uint8_t dewLimit;
    #endif
    #if defined(T_SENSOR_AVAILABLE)
    int	    ReadTemperatureX10();
    #endif
    #if defined(PROTOCOL)
    void    StopProtocol(); 
    void    ContinueProtocol();
    void    SendProtocol(); 
    void    RestartProtocol();
    bool    SendMoreProtocol();
    ProtocolStack protocol;
    #endif
    #ifdef COOLER_PWM
    float   KpKiKd[3];
    int	    PID_delay;
    int	    PID_setTemp;
    int	    PID_svdTemp;
    int     PID_error;
    bool    PID_coolingOff;
    float   PID_i;

    # define RSCAL 1	      // additional scaling of ReadTemperature() for PID calibration, 1 or 4
    # if (RSCAL==4)
    # define ReadTemperature_xS	ReadTemperatureX40
    const int initialResponseDelta_xS = 2; // 0.05 °C Schritte für die Kalibrierung mit T*40
    # else // RSCAL = 1
    # define ReadTemperature_xS	ReadTemperatureX10
    const int initialResponseDelta_xS = 1; // 0.1 °C Schritte für die Kalibrierung mit T*10
    # endif
    
    # if defined(CALIB_PID)
    StepResponse<100, RSCAL>* PIDCal;
    int	    calibState;
    int	    PID_calCnt;
    # endif

    uint32_t PID_prevTime;
    # define PID_EEPROM_ADDR	0	// address of PID coeffs saved in EEPROM
    # define SETTINGS_EEPROM_ADDR (PID_EEPROM_ADDR+4+sizeof(IntervallTimer::KpKiKd))	// address of settings saved in EEPROM
    #else
    # define SETTINGS_EEPROM_ADDR 0	// address of settings saved in EEPROM
    #endif

  public:
    #ifdef COOLER_PWM
    static int GetPIDEEPROMAddress()	    {return PID_EEPROM_ADDR;}
    #endif
    static int GetSettingsEEPROMAddress()   {return SETTINGS_EEPROM_ADDR;}

    enum {
      stIdle,
      stShoot,
      stShootPaused,
      stShootBulb,
    };
};


// Construct intervall timer.
// warnOnShootMissing enables displaying warnings on missing shoot events.
// 
inline
IntervallTimer::IntervallTimer(bool warnOnShootMissing)
{
#ifdef COOLER_PWM
  #if defined(CALIB_PID)
  PIDCal = 0;
  #endif
  #if defined(DEW_AVAILABLE)
  coolingMode = tmfAuto;		// dew control on
  #endif
#endif
  SetShootLine(false);		// (mit digitalWriteFast() setzen!)
  pinMode(SHOOT_LINE, OUTPUT);
  warnShootMissing = warnOnShootMissing;
}

inline
IntervallTimer::~IntervallTimer()
{
#ifdef COOLER_PWM
  DeletePID();
#endif
}


//
// Initialisierung, die beim Erzeugen eines globalen Objekts noch
//  nicht durchgeführt werden kann.
//
inline 
void IntervallTimer::Setup() 
{  
#if defined(SPIFLASH_CS)	          // indicates SPI flash module available
  setupSPIFlash();
#endif
  blueSerial.begin(blueBaud);	  // Serial wird erst nach globalen Objekten eingerichtet!
  setupTemperatureControl();
  Reset();			  // Kühlung erst jetzt verfügbar
#ifdef HELLO_MSG
  bluePrint_P(PSTR(HELLO_MSG));
  #if defined(SPIFLASH_CS)
    bluePrint((int)(bootFlash.getCapacity() >> 10));
    bluePrint_P(PSTR("KB ext. Flash."));
  #endif
  bluePrintln();
#endif
}


// Nächstes gepuffertes Bluetooth-Zeichen senden. Wenn Puffer leer ist,
//  zurückgestellte Aktionen weiterführen (Hilfe, PID-Calib).
// 
// Muss von der globalen Ablaufsteuerung regelmäßig aufgerufen werden!
// 
// Gibt false zurück, wenn nichts zu tun ist.
// 
inline 
bool IntervallTimer::SendNextBlue() 
{
  #if BLUESERIAL_IN_SOFTWARE
  if (blueSerial.SendNextChar()) {    // nächstes Zeichen senden
    return true;		      // gesendet
  }
  #endif
  //
  // Puffer war leer: ist was zurückgestellt?
  // 
  #if defined(PROTOCOL)
  if (SendMoreProtocol()) {	      // nächste Protokollzeile in den Sendepuffer
    return true;
  }
  #endif
  if (helpPage != 0) {
    helpPage = SendHelp(helpPage);    // nächste Charge an Zeichen in den Sendepuffer
    return true;
  }
  #if defined(CALIB_PID)
  if (calibState != 0) {
    ContinueCalib();		      // nächste Charge an Zeichen in den Sendepuffer
    return true;
  }
  #endif
  return false;			      // alles erledigt
}


// Endezeitpunkt der Auslösungsserie berechnen und eintragen
inline 
void IntervallTimer::SetShootLimit()
{
  shootLimit = GetLimit();
  if (shootLimit != 0) {	    // nicht endlos?
    shootLimit += millis();	    // Endzeit
    if (!HasShootLimit()) {
      shootLimit = 2;		    // 1 und 2 sind reserv.
    }
  }
}

// Flag für Ende der Auslösungsserie setzen oder wieder löschen.
// Rückgabe: true, wenn gesetzt
inline 
bool IntervallTimer::RequestSeriesTermination() 
{
  if (shootLimit == 1) {	    // schon gesetzt?
    SetShootLimit();		    //  dann Limit neu herstellen
    return false;
  }
  else {
    shootLimit = 1;
    return true;
  }
}

// Zeitbegrenzte Serie?
inline 
bool IntervallTimer::HasShootLimit() const
{
  return shootLimit >= 2;
}

// Auslösung (de-)aktivieren
inline 
void IntervallTimer::SetShootLine(bool active) 
{
  digitalWriteFast(SHOOT_LINE, active ? 0 : 1);
}


// 
// Neue zufällige Abweichung von der Belichtungszeit festlegen
//
inline 
void IntervallTimer::RandomSpeedOffset()
{
  // GetSpeed() * random(speedDither / 100) mit besserer Auflösung:
  speedOffset = ((GetSpeed() >> 6)
		 * random(speedDither * ((1<<12)/100))	    // vorab schon durch 100 teilen
		) >> 6;
}


// Auslösesignal anlegen und Status auf stShoot setzen.
// In endShoot die Zeit für das Ende der Auslösung eintragen.
// 
// Achtung: Die vorherige Auslösung ist noch nicht angezeigt/protokolliert.
//  Deshalb später numFrames - 1 verwenden.
//
inline 
void IntervallTimer::StartFrame()
{
  if (!IsFastSpeed()) {			// (für FastSpeed nicht, sonst wird Flag vor Nutzung überschrieben!)
    shootHappend = false;		// Emulator hat noch nicht reagiert
  }
  SetShootLine(true);			//   los!
  bgnShoot = millis();
  endShoot = bgnShoot + GetSpeed();
  if (speedIndex != 0  &&  speedDither > 0) {	// Bulb dithern?
    endShoot += speedOffset;		// Offset zugeben
    if (speedOffset > 0) {		// nächstes mal negativ
      speedOffset = -speedOffset;
    }
    else {
      RandomSpeedOffset();
    }
  }
  shootState = stShoot;
  nextUpdate = -1;			// Start anzeigen, wird für Folgeauslösungen noch auf -2 gesetzt
  numFrames++;
}


///////////////////////////////////////////////////////////////
inline bool timeReached(uint32_t t) __attribute__((always_inline));
inline bool timeReached(uint32_t t, uint32_t now) __attribute__((always_inline));

// Prüfen, ob der Zeitpunkt t (in msec) erreicht ist
inline bool timeReached(uint32_t t)
{
  return timeReached(t, millis());
}

// Prüfen, ob der Zeitpunkt t (in msec) now erreicht hat.
inline bool timeReached(uint32_t t, uint32_t now)
{
  return (int32_t)(t - now) < 0;	// auch ok für Überlauf von milliTime bis 0x7FFFFFFF
}


// Prüfen, ob zum aktuellen Auslöse-Zustand Aktionen erforderlich sind.
// - Zeit ist rum: Pause
// - Pause zu Ende: nächste Auslösung
//
inline 
void IntervallTimer::Advance()
{
  if (shootState == stShootPaused) {	// beendete Auslösung:
    if (timeReached(endShoot)) {	// Pause zu Ende?
      StartFrame();
      nextUpdate = -2;			// Folge-Auslösung
    }
  }
  else if (shootState == stShoot) {     // laufende Auslösung: 
    if (timeReached(endShoot)) {	// Ende erreicht?
      //
      // Auslösesignal beenden und Status auf stShootPaused setzen.
      // In endShoot die Zeit für die nächste Auslösung (nach delay) eintragen.
      //
      SetShootLine(false);		// aus
      shootState = stShootPaused;	// Status erts mal auf Pause
      endShoot = millis() + (IsFastSpeed() ? shootBreak_ms : bulbBreak_ms); // Pausenende eintragen
      CheckShootLimit(endShoot);	// ganz abschalten, wenn Serie erledigt
    }
  }
}

//
// Prüfen, ob zur Zeit 'when' das Shoot-Limit erreicht ist und ggf. abschalten
// 
// Rückgabe: true, wenn Limit erreicht ist
//
inline 
bool IntervallTimer::CheckShootLimit(uint32_t when)
{
  if (HasShootLimit()  &&  timeReached(shootLimit, when)) {	// Zeitlimit erreicht?
    shootState = stIdle;	// fertig
    PrintFrameSummary(0);	// Abschluss des Frames (wird sonst immer zu Beginn des nächsten angezeigt)
    bluePrintln_P(PSTR("...Limit"));
    #if defined(COOLER_PWM) || defined(COOLER_POWER)
    if (IsCooling() || !IsPIDOff()) {
      AbortPID();
      SelectCoolingMode(0);
      bluePrintln_P(PSTR("(Kühlung abgeschaltet)"));
    }
    #endif
    return true;
  }
  if (shootLimit == 1) {	// Serie abbrechen?
    shootState = stIdle;
    PrintFrameSummary(0);	// Abschluss des Frames (wird sonst immer zu Beginn des nächsten angezeigt)
    return true;
  }
  return false;
}

//
// Info zum Abschluss des Frames (numFrames + frameOfs) ausgeben,
//  ggf. Protokoll eintragen.
// 
inline 
void IntervallTimer::PrintFrameSummary(int frameOfs)
{
  bluePrint_P(PSTR("* ("));	  //   zum letzten Frame noch Count und Temp. anzeigen
  blueSerial.print(numFrames + frameOfs);

  #if defined(T_SENSOR_AVAILABLE) || defined(DEW_AVAILABLE)	  // Temp-Sensor vorhanden:
    bluePrint_P(PSTR(": "));
    #if defined(T_SENSOR_AVAILABLE)
      int Ts = ReadTemperatureX10();
      bluePrintTdegC(Ts);
    #endif
    //
    // Umgebungstemperatur lesen:
    // (Einen I2C Sensor wie den AHT20 auszulesen dauert zu lange, wenn
    //  die nächste Auslösung schon gestartet ist. Ohne die angepasste
    //  Treiberversion (AHT20C) müsste die Temperatur ggf. schon während
    //  der Aufnahme gelesen und in Tu hinterlegt werden.)
    //
    int	Tu;
    if (ReadAmbientTemperatureX10(Tu)) {
      bluePrint_P(PSTR(", Tu="));
      bluePrintTdegC(Tu);
    }
  #endif
  bluePrint_P(PSTR(")"));
  
  #if defined(PROTOCOL)	
    // Für das Protokoll nicht auf Vorhandensein von Tu prüfen, weil das 
    // Rücklesen nicht erkennen kann ob Tu im Protokoll enthalten ist!
    //
    #if defined(PROTOCOL_2T)
      protocol.Add(numFrames + frameOfs, Ts, Tu);
    #elif defined(T_SENSOR_AVAILABLE) 
      protocol.Add(numFrames + frameOfs, Ts);
    #elif defined(DEW_AVAILABLE)
      protocol.Add(numFrames + frameOfs, Tu);
    #endif
  #endif
}

//
// Info über Bluetooth senden
// 
// Rückgabe: true, wenn alles erledigt ist
//
inline 
bool IntervallTimer::SendUpdate()
{
  uint32_t now = millis();
  if (shootState == stShootBulb) {	// Dauerauslösung:
    if (!quiet) {
      if (timeReached(nextUpdate, now)) {
	blueSerial.write(' ');
	blueSerial.print((int)((now - endShoot) / 1000)); // abgelaufene Sekunden anzeigen
	do {
	  nextUpdate += 1000;
	} while (timeReached(nextUpdate, now));  // weiter bis aktuelle Zeit erreicht (falls zuvor quiet!)
      }
    }
  }
  else while (shootState == stShoot) { // aktiv?
    static uint32_t deltaStern;
    static uint8_t deltaCnt;
    static uint8_t deltaWarnShoot;
    if (nextUpdate < 0) {		  // neue Auslösung?
      if (IsFastSpeed()) {		  // schnelle Auslösungen:
	//
	// Erste Auslösung der Reihe oder 1s vergangen?
	// (Zeit separat in deltaStern abgelegt, weil nextUpdate durch
	//  nächste Auslösung immer -1 wird!)
	//
	if (nextUpdate == -1 || timeReached(deltaStern, now)) {  
	  if (!quiet) {
	    if (shootHappend) {		  // Coil-Signal empfangen seit dem letzten Auslösen?
	      bluePrint_P(PSTR("'"));	  // Auslösung anzeigen
	      shootHappend = false;	  // wir müssen selbst zurücksetzen!
	    }
	    else {
	      bluePrint_P(PSTR("."));
	      StopCoolingIdleCam(255 - 240);  // nach 240s ggf. Kühlung abschalten
	    }
	  }
	  deltaStern = now + 1000;		
	}
      }
      else {				  // Neue Bulb-Auslösung:
	if (!quiet) {
	  if (nextUpdate == -2) {	  //  weitere Auslösung einer Reihe?
	    PrintFrameSummary(-1);	  //   Abschluss der letzten Auslösung anzeigen
	  }
	  bluePrintln();
	  bluePrint((int)(GetSpeed() / 1000));
	  bluePrint_P(PSTR("s: "));
	  deltaStern = GetSpeed() >> 4;	//  in 16(s. XX) Abschnitte unterteilen
	}
	// Bei kurzen Zeiten (< ~10s) ist die Auslösung beim ersten Stern noch 
	//  nicht passiert. Damit nicht fälschlich "Keine Auslösung" gemeldet 
	//  wird, darf die Prüfung nicht so früh erfolgen. Deshalb erst nach 
	//  4 Sternen.
	// Bei sehr kurzen Zeiten (1s) reicht auch das noch nicht. Deshalb 
	//  eigenen Counter für 0.5s einrichten: Bei jedem Stern, also nach
	//  deltaStern msec, wird gezählt, also:
	//
	if (deltaStern > 500) {
	  deltaWarnShoot = 0;	  // ab 500ms beim 1. Stern
	}
	else {
	  deltaWarnShoot = uint8_t(500 / (uint8_t)deltaStern);
	}
	deltaCnt = 0;
      }
    }
    else if (timeReached(nextUpdate, now)) {	// laufende Auslösung: ausreichend fortgeschritten
      deltaCnt++;
      if (!quiet) {
	blueSerial.write('*');		// Sterne anzeigen: 16 (+ XX, wenn in NotifyShoot() 'endShoot' korrigiert wurde!)
	if ((deltaCnt & 3) == 0) {	// "**** **** **..  "
	  blueSerial.write(' ');
	}
      }
      if (deltaWarnShoot == 0) {
	if (!shootHappend) {		// Auslösung nicht gemeldet:
	  if (warnShootMissing) {
	    bluePrint_P(PSTR(" Keine Auslösung!? "));
	  }
	  shootHappend = true;		// nur einmal melden!
	  StopCoolingIdleCam(255-3);	// nach 3 Frames ohne Trigger ggf. Kühlung abschalten
	}
      }
      else {
	deltaWarnShoot--;
      }
      if (deltaCnt == 11) {
	TriggerAmbientSensor();		// neue Messung auslösen für den Fall, dass Zykluszeit größer als BeliZeit
      }
    }
     else {				// noch warten
      break;
    }
    //
    // Zeit für nächstes Update = jetzt + 1s
    //
    if (IsFastSpeed()) {		// Kamera nicht auf Bulb
      // weiteres Update verhindern, bis nextUpdate beim 
      // Auslösen wieder = -1 gesetzt wird, oder bis 1s vorbei:
      // 
      nextUpdate = now + 1000;		
    }
    else {				// lange Zeiten:
      nextUpdate = now + deltaStern;	//  nächster der 16 Abschnitte
    }
    break;
  }
  return shootState == stIdle;		// fertig, wenn 0 erreicht
}


//
// Benachrichtigung über eine Auslösung durch den Shutteremulator entgegennehmen
// 
inline 
void IntervallTimer::NotifyShoot()
{
  shootHappend = true;
  shootOnceHappend = 255;		// Flag: Shutter hat schon mal gemeldet
  if (shootState == stShoot) {		// bei laufender Auslösung
    endShoot += millis() - bgnShoot;	//  Verzögerung durch Shutter und Speichern korrigieren
  }
}

//
// Wenn der Shutter schon mal gemeldet hat und dann nicht mehr,
//  sollten wir ggf. die Kühlung abschalten, weil die Kamera sonst
//  deutlich kälter wird als geplant und beschlagen könnte.
//
// Dazu wird in shootOnceHappend 255 hinterlegt, wovon bis auf 'when'
//  heruntergezählt wird.
//
inline
void IntervallTimer::StopCoolingIdleCam(uint8_t when)
{
#if defined(COOLER_PWM) || defined(COOLER_POWER)
  if (IsCooling() && shootOnceHappend != 0) {	// Shutter hat schon gemeldet und die Kühlung läuft noch:
    shootOnceHappend--;		// nach 3 (s. NotifyShoot()) Frames
    if (shootOnceHappend == when) {	//  abschalten:
      SelectCoolingMode(0);
      PID_coolingOff = false;	//   und verhindern, dass PID wieder aktiviert
      bluePrint_P(PSTR("Kühlung abgeschaltet!"));
    }
  }
#else
  when = when;
#endif
}

#if defined(PROTOCOL)

inline
void IntervallTimer::StopProtocol()
{
  protocol.Stop();
  bluePrintln_P(PSTR("Aufzeichnung angehalten."));
}

inline
void IntervallTimer::ContinueProtocol()
{
  protocol.Continue();
  bluePrintln_P(PSTR("Aufzeichnung wird fortgesetzt."));
}

// Aufzeichnung anzeigen bzw. beenden, wenn pausiert.
inline
void IntervallTimer::SendProtocol()
{
  if (protocol.IsSendingPaused()) {
    protocol.StopSending();
  }
  else {
    if (!protocol.StartSending()) {
      bluePrintln_P(PSTR("Keine Aufzeichnung vorhanden!"));
      return;
    }
  }
  return;     // Ausgabe wird vom Emu aufgerufen
}

//
// Nächste Protokollzeile an Blue geben.
//
// Gibt false zurück, wenn nichts mehr zu senden war oder eine Pause ansteht.
//
inline
bool IntervallTimer::SendMoreProtocol()
{
  if (!protocol.IsSending() || protocol.IsSendingPaused()) {
    return false;
  }
  if (blueGetFreeCount() < blueGetHalfSize()) {
    return false;
  }
  if (!protocol.ReadNextLine(8)) {
    return false;
  }

  bluePrint(protocol.GetRead_I());
  bluePrint_P(PSTR(": "));
  #if defined(T_SENSOR_AVAILABLE)
    bluePrintTdegC(protocol.GetRead_T1());
  #endif
  #if defined(DEW_AVAILABLE)
    #if defined(PROTOCOL_2T)	// auch Sensor:
      bluePrint_P(PSTR(", Tu="));
      bluePrintTdegC(protocol.GetRead_T2());
    #else
      bluePrint_P(PSTR("Tu="));
      bluePrintTdegC(protocol.GetRead_T1());
    #endif
  #endif
  bluePrintln();
  
  return true;
}


inline
void IntervallTimer::RestartProtocol()
{
  protocol.Reset();
}
#endif

#endif
