/////////////////////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//
//       17.05.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////

#include "IntervallTimer.h"

#include <EEPROM.h>
#include <avr/wdt.h>


//
// Reset/initialise IntervallTimer and load settings from EEPROM if avl. or basic value set.
//
void IntervallTimer::Reset() 
{  
  helpPage = 0;
  sentNewline = 0;
  numDigits = 0;
  inputMode = 0;
  shootState = stIdle;
  numFrames = 0;
  SetShootLine(false);
#ifdef COOLER_PWM
  AbortPID();
  ResetPID(true);
#endif
  #if defined(CALIB_PID)
  calibState = 0;
  #endif

  ResetSettings(true);	// load settings from EEPROM (if avl)

  /*
#include <SPI.h>
#define READ 0x03
#define WRITE 0x02
#define WREN 0x06  // write enable
  SPI.begin();
  SPI.transfer(WREN);
  SPI.transfer(WRITE);
  int addr = 1000;
  int value = 12;
  SPI.transfer((byte)(addr>>8));
  SPI.transfer((byte)(addr&0xFF));
  SPI.transfer((byte)(value>>8));
  SPI.transfer((byte)(value&0xFF));
  */
}

//
// Echten Reset per Watchdog durchführen
// 
// Achtung: Klappt nicht auf Nano und Pro Mini mit dem 
//	    originalen Bootloader! (urboot ist ok.)
//
void IntervallTimer::watchdogReset()
{
  /*
  cli(); // disable all interrupts
  asm("WDR");
  WDTCSR |= (1<<WDCE) | (1<<WDE);   // Watchdog: allow changes, disable reset
  WDTCSR = (1<<WDIE) | (1<<WDE);    // 16ms / interrupt, system reset
  sei();
  /*/ //oder:
  //#include <avr/wdt.h>
  wdt_enable(WDTO_15MS); // 16ms / System Reset
  //WDTCSR = (1<<WDIE); // interrupt
  //*/
}

//
// Load settings from EEPROM (if avl && fromEEPROM==true) or basic value set.
//
// Returns true if loaded from EEPROM
//
bool IntervallTimer::ResetSettings(bool fromEEPROM)
{
  if (fromEEPROM && LoadSettingsfromEEPROM()) {
    WriteCoolingLevel();
    WriteHeatingLevel();
    return true;
  }

  // Basis-Einstellungen
  //
  speedIndex = 2;     // 60s
  limitIndex = 2;     // 12h
  quiet = false;
  debug = false;
  bulbBreak_ms = bulbBreak_min;   // Pause zwischen Bulb-Auslösungen, Minimalwert
  speedDither = 0;
  tempCorr = 0
    #ifdef Ts_CORRECTION                   // predefined corrective addition to the sensor temperature
    + int(Ts_CORRECTION * 10)
    #endif
    ;
  TuCorr = 0
    #ifdef Tu_CORRECTION                   // predefined corrective addition to the ambient temperature
    + int(Tu_CORRECTION * 10)
    #endif
    ;
  TdCorr = 0
    #ifdef Td_CORRECTION                   // predefined corrective addition to the ambient temperature
    + int(Td_CORRECTION * 10)
    #endif
    ;
#if defined(COOLER_PWM) || defined(COOLER_POWER)
  SelectCoolingMode(-1);
#endif
#if defined(HEATER_POWER) || defined(HEATER_PWM)
  SelectHeatingMode(-1);
#endif 
  return false;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Belichtungszeiten in Millisekunden
//
const uint32_t speedTbl[] = {100, 30 sec, 60 sec, 100 sec, 180 sec, 300 sec, 600 sec, 1200 sec, 2400 sec, 3600 sec}; // BeLi-Zeiten in sec
//        --> in Minuten:                  1                 3        5       10        20        40        60 min


// Zeitgrenzen in Millisekunden, 0 == endlos.
//
const uint32_t limitTbl[] = {0, 16*60 Min, 12*60 Min, 8*60 Min, 6*60 Min, 4*60 Min, 2*60 Min, 60 Min, 30 Min, 20 Min, 10 Min}; 


//
// Gewählte Belichtungszeit in der Basiseinheit (ms) zurückgeben
//
uint32_t IntervallTimer::GetSpeed() const
{
  if (speedIndex < 0) {
    return (-speedIndex) sec;	  // negativer Index == Zeit in Minuten
  }
  return speedTbl[speedIndex];
}

//
// Gewählte Zeitgrenze in der Basiseinheit (ms) zurückgeben
//
uint32_t IntervallTimer::GetLimit() const
{
  if (limitIndex < 0) {
    return (-limitIndex) Min;	  // negativer Index == Zeit in Minuten
  }
  return limitTbl[limitIndex];
}

//
// Neue Zeitgrenze in min eintragen und ggf. anwenden, neuen Status anzeigen
//
// Return true if value is ok (1 - MAXINT).
//
bool IntervallTimer::SetLimit(int32_t minutes) 
{
  if (minutes > 0 && minutes <= 0x7FFFL) {	  // nur positive Werte!
    limitIndex = -(int)minutes;	    // in Minuten eintragen
    SetShootLimit();		    // auf laufende Serie anwenden
  }
  else if (minutes == 0) {	    // endlos
    limitIndex = 0;
  }
  else {
    SendErrorBlue();
    return false;
  }
  SendInfoBlue(2);
  return true;
}

//
// Gewählte Zeitgrenze in der Basiseinheit (ms) zurückgeben
//
int IntervallTimer::GetNumSpeeds() const
{
  return (int)ELEMENTS(speedTbl);
}

//
// Gewählte Zeitgrenze in der Basiseinheit (ms) zurückgeben
//
int IntervallTimer::GetNumLimits() const
{
  return (int)ELEMENTS(limitTbl);
}


//////////////////////////////////////////////////////////////////////////
//
// Save settings to EEPROM
// 
// Einstellungen:
//
#define SETTINGS_EEPROM_SIZE  (4 + sizeof(ivtSettings) + sizeof(debug))

void ivtSettings::SaveSettingstoEEPROM()
{
  int iEE = IntervallTimer::GetSettingsEEPROMAddress();	  // EEPROM address
  EEPROM.update(iEE++, 'S');	  // magic
  EEPROM.update(iEE++, 'E');
  EEPROM.update(iEE++, 'T');
  EEPROM.update(iEE++, SETTINGS_EEPROM_SIZE);
  uint8_t* p = (uint8_t*)&debug;
  for (unsigned int i = 0; i < sizeof(debug); i++) {
    EEPROM.update(iEE++, *p++);
  }
  p = (uint8_t*)this;
  for (unsigned int i = 0; i < sizeof(ivtSettings); i++) {
    EEPROM.update(iEE++, *p++);
  }
}


//
// Clear settings from EEPROM
//
void ivtSettings::ClearSettingsfromEEPROM()
{
  EEPROM.update(IntervallTimer::GetSettingsEEPROMAddress(), 0);	  // clear magic only
}


//
// Load settings from EEPROM
// 
// Returns false if not avl.
//
bool ivtSettings::LoadSettingsfromEEPROM()
{
  int iEE = IntervallTimer::GetSettingsEEPROMAddress();	  // EEPROM address
  char c[4];
  EEPROM.get(iEE++, c[0]);		  // check magic (survives scetch uploads)
  EEPROM.get(iEE++, c[1]);
  EEPROM.get(iEE++, c[2]);
  EEPROM.get(iEE++, c[3]);	    // data size
  if (c[0] == 'S' && c[1] == 'E' && c[2] == 'T' && c[3] == SETTINGS_EEPROM_SIZE) {
    uint8_t* p = (uint8_t*)&debug;
    for (unsigned int i = 0; i < sizeof(debug); i++) {
      EEPROM.get(iEE++, *p++);
    }
    p = (uint8_t*)this;
    for (unsigned int i = 0; i < sizeof(ivtSettings); i++) {
      EEPROM.get(iEE++, *p++);
    }
    return true;
  }
  return false;
}



#if defined(COOLER_PWM)
//
// Save PID coeffs to EEPROM
//
void IntervallTimer::SavePIDtoEEPROM()
{
  int iEE = GetPIDEEPROMAddress();	  // EEPROM address
  EEPROM.update(iEE++, 'P');	  // magic
  EEPROM.update(iEE++, 'I');
  EEPROM.update(iEE++, 'D');
  EEPROM.update(iEE++, sizeof(KpKiKd));
  uint8_t* p = (uint8_t*)KpKiKd;
  for (unsigned int i = 0; i < sizeof(KpKiKd); i++) {
    EEPROM.update(iEE++, *p++);
  }
}


//
// Clear PID coeffs from EEPROM
//
void IntervallTimer::ClearPIDfromEEPROM()
{
  EEPROM.update(GetPIDEEPROMAddress(), 0);	  // clear magic only
}


//
// Load PID coeffs from EEPROM
// 
// Returns false if not avl.
//
bool IntervallTimer::LoadPIDfromEEPROM()
{
  int iEE = GetPIDEEPROMAddress();	  // EEPROM address
  char c[4];
  EEPROM.get(iEE++, c[0]);		  // check magic (survives scetch uploads)
  EEPROM.get(iEE++, c[1]);
  EEPROM.get(iEE++, c[2]);
  EEPROM.get(iEE++, c[3]);	    // data size
  if (c[0] == 'P' && c[1] == 'I' && c[2] == 'D' && c[3] == sizeof(KpKiKd)) {
    uint8_t* p = (uint8_t*)KpKiKd;
    for (unsigned int i = 0; i < sizeof(KpKiKd); i++) {
      EEPROM.get(iEE++, *p++);
    }
    return true;
  }
  return false;
}

#endif

//////////////////////////////////////////////////////////////////////////
// Fehler über Bluetooth melden
//
void IntervallTimer::SendErrorBlue()
{
  EnterNewLine(PSTR("Ungültige Eingabe!"));
}


//
// Frische Zeile anfangen und s ausgeben.
//
void IntervallTimer::EnterNewLine(PGM_P s)
{
  EnterNewLine();
  bluePrint_P(s);
}


//
// Frische Zeile anfangen und s ausgeben und neue Zeile anfangen.
//
void IntervallTimer::EnterNewLineLn(PGM_P s)
{
  EnterNewLine();
  bluePrint_P(s);
  PrintNewLine();
}


//
// Frische Zeile anfangen, wenn das BT-Terminal keine neue Zeile registriert hat.
//
void IntervallTimer::EnterNewLine()
{
  if (sentNewline != 0) {	  // wurde vor (!) dem aktuellen Ereignis LF gesendet?
    sentNewline = 0;		  //  dann verbraucht
  }
  else {
    bluePrintln();	    // BT-Terminal sendet kein LF: dann neue Zeile
  }
}


//
// Zeilenumbruch ausgeben.
//
void IntervallTimer::PrintNewLine()
{
  bluePrintln();
  sentNewline = 1;
}


//
// " (" oder ", " anzeigen
//
void IntervallTimer::PrintMore(bool more)
{
  if (more) {
    bluePrint_P(PSTR(", "));
  }
  else {
    bluePrint_P(PSTR(" ("));
  }
}


//
// PMW Stauts anzeigen
//
void IntervallTimer::PrintCoolerStatus()
{
#if defined(COOLER_PWM) || defined(COOLER_POWER)
  bluePrint_P(PSTR("PWM="));
  if (IsCooling()) {
    blueSerial.print(coolingLevel);
  }
  else {
    bluePrint_P(PSTR("---"));
  }
  #if defined(DEW_AVAILABLE)
  if (IsDewControl()) {	    // Kühlung an Taupunkt anpassen?
    bluePrint_P(PSTR("~")); 
  }
  #endif
#endif
}


//
// PMW Stauts anzeigen
//
void IntervallTimer::PrintHeaterStatus()
{
#ifdef HEATER_POWER
  bluePrint_P(PSTR("Heizung "));
  if (IsHeating()) {	        // an:
    #if HEATER_POWER_BY_COOLER	  // no heating if cooler and fan power are off
    if (!IsCooling()) {
      bluePrint_P(PSTR("..."));	      //  auto-off
    }
    else 
    #endif
    {
    #if defined(HEATER_SOFT_PWM) || defined(HEATER_PWM)
      // zur linearen Anzeige müsste der mode gespeichert werden oder sqrt(heatingLevel):
      blueSerial.print(((heatingLevel+1)*100)>>8);
      bluePrint_P(PSTR("%"));
    #else
      bluePrint_P(PSTR("an"));
    #endif
    }
    #if defined(DEW_AVAILABLE)
    if (IsDewHeating()) {	    // Heizung an Taupunkt anpassen?
      bluePrint_P(PSTR("~"));
    }
    #endif
  }
  else {
    bluePrint_P(PSTR("aus"));
  }
#endif
}


//
// PID-Status anzeigen
//
#if defined(COOLER_PWM)
void IntervallTimer::PrintPIDTemp()
{
  bluePrint_P(PSTR("Temperaturregelung: Ziel = "));
  bluePrintTdegC(PID_setTemp - tempCorr);
  #if defined(DEW_AVAILABLE)
  if (IsDewControl()) {
    bluePrint_P(PSTR(" >= Taupunkt"));
  }
  #endif
}
#endif


//
// Speed anzeigen
//
void IntervallTimer::PrintSpeed()
{
  if (IsFastSpeed()) {
    bluePrint_P(PSTR("Schnelle Auslösungen für kameraseitig gewählte Zeit"));
  }
  else {
    bluePrint_P(PSTR("Bulb-Belichtungszeit: "));
    blueSerial.print((int)(GetSpeed() / 1000));
    bluePrint_P(PSTR("s"));
  }
}


//
// Info zu Einstellungen über Bluetooth senden, which = Bitmaske:
// -1: alle
// 1: Speed
// 2: Serien-Limit
//
void IntervallTimer::ShowSettings()
{
  EnterNewLine();		  // frische Zeile
  PrintSpeed();			  // Belichtungszeit
  bluePrint_P(PSTR(" +/- "));
  blueSerial.print(speedDither);  // Dither
  bluePrint_P(PSTR("%"));
  bluePrint_P(PSTR(" + "));
  int secs = (int)(bulbBreak_ms / 1000);	//  Sekunden
  blueSerial.print(secs);
  int msecs = (int)(bulbBreak_ms - (int32_t)secs * 1000);	// restliche millis
  if (msecs >= 100) {			//  + ggf. Zehntel
    bluePrint_P(PSTR("."));
    blueSerial.print(msecs / 100);
  }
  bluePrint_P(PSTR("s Pause"));
  SendInfoBlue(2);	  // Limit

  #if defined(COOLER_PWM)
    if (IsPIDActive()) {
      PrintPIDTemp();
    }
    else
    #endif
    #if defined(COOLER_PWM) || defined(COOLER_POWER)
    PrintCoolerStatus();
  #endif
  #ifdef HEATER_POWER
  bluePrint_P(PSTR(", "));
  PrintHeaterStatus();
  bluePrintln();
  #endif

  bluePrint_P(PSTR("Updates "));
  if (quiet) {
    bluePrint_P(PSTR("aus"));
  }
  else {
    bluePrint_P(PSTR("aktiv"));
  }
  bluePrint_P(PSTR(", Debug-Infos a"));
  if (debug) {
    bluePrintln_P(PSTR("n"));
  }
  else {
    bluePrintln_P(PSTR("us"));
  }
  StartHelp(settingsPart);			// wegen Bluetooth-Überlauf später weiter
}
  

//
// Info zu Einstellungen, Teil 2:
// 
int IntervallTimer::ShowSettingsPart2()
{
  bool tc = false;
  #if defined(T_SENSOR_AVAILABLE)
  tc = (tempCorr != 0);
  #endif
  #if defined(DEW_AVAILABLE)
  tc = tc || (TuCorr != 0) || (TdCorr != 0);
  #endif
  if (tc) {
    bluePrint_P(PSTR("Temperatur-Korrektur "));
    #if defined(T_SENSOR_AVAILABLE)
    bluePrintTdegCsign(tempCorr);
    #endif
    #if defined(DEW_AVAILABLE)
    bluePrint_P(PSTR("   Tu: "));
    bluePrintTdegCsign(TuCorr);
    bluePrint_P(PSTR("   Td: "));
    bluePrintTdegCsign(TdCorr);
    #endif
    bluePrintln();
  }
  return 0; // fertig
}

//
// Info über Bluetooth senden, which = Bitmaske:
// -1: alle
// 1: Speed
// 2: Serien-Limit
//
void IntervallTimer::SendInfoBlue(int which)
{
  EnterNewLine();		  // frische Zeile
  if (which & 1) {
    PrintSpeed();
    if (!IsFastSpeed()) {
      if (speedDither != 0) {
	bluePrint_P(PSTR(" +/- "));
	blueSerial.print(speedDither);
	bluePrint_P(PSTR("%"));
      }
      if (bulbBreak_ms != bulbBreak_min) {	// mit zusätzlicher Pause?
	bluePrint_P(PSTR(" + "));
	int secs = (int)(bulbBreak_ms / 1000);	//  Sekunden
	blueSerial.print(secs);
	int msecs = (int)(bulbBreak_ms - (int32_t)secs * 1000);	// restliche millis
	if (msecs >= 100) {			//  + ggf. Zehntel
	  bluePrint_P(PSTR("."));
	  blueSerial.print(msecs/100);
	}
	bluePrint_P(PSTR("s Pause"));
      }
    }
    int more = 0;
    #ifdef T_SENSOR_AVAILABLE
      if (IsFastSpeed()) {
        bluePrintln();
      }
      bluePrint_P(PSTR(" ("));
      /* 10 mal mitteln: Nur für Test, dauert zu lange!
      int meanResponse = ReadTemperatureX10();      
      int meanCnt = 0;
      while (meanCnt < 9) {
	meanCnt++;
	meanResponse += ReadTemperatureX10();      // take the mean of 10 10x responses
      }
      bluePrintDecimals(meanResponse, 2);
      */
      bluePrintDecimals(ReadTemperatureX10(), 1);
      #if defined(COOLER_PWM)
      if (IsPIDActive()) {
	bluePrint_P(PSTR("->"));
	bluePrintDecimals(PID_setTemp - tempCorr, 1);
      }
      #endif
      bluePrint_P(PSTR("°C"));
      more = 1;
    #endif // T_SENSOR_AVAILABLE
    #if defined(COOLER_PWM) || defined(COOLER_POWER)
      PrintMore(more);
      PrintCoolerStatus();
      more = 1;
    #endif
    #ifdef HEATER_POWER
      PrintMore(more);
      PrintHeaterStatus();
      more = 1;
    #endif
    int	Tu;
    if (ReadAmbientTemperatureX10(Tu)) {
      bluePrint_P(PSTR(", Tu="));
      bluePrintTdegC(Tu);
    }
    float humidity;
    if (ReadAmbientHumidity(humidity)) {
      bluePrint_P(PSTR(", RF="));
      bluePrint(int(humidity));
      bluePrint_P(PSTR("%, Td="));
      int td = CalcDewPointX10(humidity, Tu);
      if (TdCorr != 0) {		    // "Td=xx-yy°C"
        bluePrintDecimals(td, 1);
        bluePrintTdegCsign(-TdCorr);
      }
      else {
	bluePrintTdegC(td);
      }
    }
    TriggerAmbientSensor();		// neue Messung auslösen für den Fall, dass kein Shutter emuliert wird
    if (more != 0) {
      bluePrint_P(PSTR(")"));
    }
    bluePrintln();
  }
  if (which & 2) {
    if (GetLimit() == 0) {
      bluePrintln_P(PSTR("(Kein Limit)"));
    }
    else {
      bluePrint_P(PSTR("Ende nach "));
      int hrs = (int)(GetLimit() / 3600000);
      blueSerial.print(hrs);
      bluePrint_P(PSTR("h "));
      int mins = (int)(GetLimit() / 60000) - hrs * 60;
      if (mins > 0) {
	blueSerial.print(mins);
	bluePrint_P(PSTR("min "));
      }
      if (!IsFastSpeed()) {
	bluePrint_P(PSTR("bzw. "));
	uint32_t n = GetLimit() / (GetSpeed() + bulbBreak_ms);
	blueSerial.print(n);
	if ((which & 1) == 0) {		  // falls nur Limit: 
	  bluePrint_P(PSTR(" Frames a "));	  //  auch Frame-Gesamtzahl anzeigen
	  blueSerial.print((int)(GetSpeed() / 1000));
	  bluePrintln_P(PSTR("s"));
	}
	else {
	  bluePrintln_P(PSTR(" Frames"));
	}
      }
      else {
	bluePrintln();
      }
    }
  }
  if (which < 0) {	    // aktuellen Status anzeigen:
    bluePrint_P(PSTR("Status: "));
    if (shootState == stShootBulb) {	// Bulb-Auslösung
      bluePrint_P(PSTR("Bulb ("));
      blueSerial.print((int)((millis() - endShoot)/1000));
      bluePrint_P(PSTR("s, ...)"));
    }
    else if (shootState == stIdle) {	// tut nix
      bluePrint_P(PSTR("Bereit"));
    }
    else {				// Serie
      if (!IsFastSpeed()) {
	blueSerial.print(numFrames-1);  // Frames bisher
	if (HasShootLimit()) {
	  bluePrint_P(PSTR(" von "));
	  // Gesamt = T/t, wobei t abweichen kann von früheren Werten
	  // Also nicht T/t - n!
	  //
	  long rest = shootLimit - millis();  // Restzeit
	  if (rest <= 0) {
	    blueSerial.print(numFrames);	// letzter läuft
	  }
	  else {
	    blueSerial.print((int)(rest / GetSpeed()) + (numFrames - 1));
	  }
	}
	bluePrint_P(PSTR(" Frames, + "));
	if (shootState == stShootPaused) {
          blueSerial.print((int)(GetSpeed()/1000));
	}
	else {
          blueSerial.print((int)((GetSpeed() - (endShoot - millis()))/1000));
	}
	bluePrint_P(PSTR(" von "));
        blueSerial.print((int)(GetSpeed()/1000));
	bluePrint_P(PSTR("s"));
      }
      else {
        bluePrint_P(PSTR("Schnelle Auslösung .."));
      }
    }
    bool xtra = false;
    if (quiet) {
      bluePrint_P(PSTR(" (keine Updates"));
      xtra = true;
    }
#if defined(PROTOCOL)
    PrintMore(xtra);
    xtra = true;
    if (protocol.IsActive()) {
      bluePrint(protocol.GetCount());
      bluePrint_P(PSTR(" aufgezeichnet"));
    }
    else {
      bluePrint_P(PSTR("Aufzeichnung aus"));
    }
#endif
    if (debug) {
      PrintMore(xtra);
      bluePrint_P(PSTR("Debug-Infos an"));
      xtra = true;
    }
    if (xtra) {
      bluePrintln_P(PSTR(")"));
    }
    else {
      PrintNewLine();
    }
  }
  // Speedtest:
  //blueSerial.println((int)(millis()-mm)); // --> 91 ms für ca. 80 Zeichen, kein Puffer, blockiert alles!!!
  // 2 ms mit BufferedSoftwareSerial!
}

//
// Hilfe über Bluetooth senden.
// Um den Sendepuffer klein zu halten, wird die Hilfe in mehrere
//  Teile zerlegt, die jeweils erst gesendet werden, wenn der
//  Sendepuffer wieder geleert ist.
// 'what' = 1...n zeigt an, welcher Teil ansteht.
/*
Der Intervalltimer kann die Kamera periodisch auslösen. Für Langzeit-Belichtungen wird an der Kamera der 
Bulb-Mode gewählt und der Intervalltimer bestimmt die Belichtungszeit.
Wenn die Belichtungszeit von der Kamera vorgegeben werden soll, wird der Intervalltimer auf die minimale
Belichtungszeit von 0.1 Sekunden eingestellt und löst dann zehn mal pro Sekunde aus.

Die Wahl der Betriebsparameter erfolgt über die zuvor hergestellte Bluetooth-Verbindung durch Eingabe einzelner Befehlszeichen, die sofort ausgeführt werden.
Einige Befehle erwarten eine Ziffernfolge, die vor dem Befehlszeichen eingegeben wird.
Der Intervalltimer verliert seine Einstellungen, wenn die Kamera ausgeschaltet wird.
...
Bei bestehender Bluetooth-Verbindung wird der aktuelle Belichtungsstatus kontinuierlich angezeigt (sofern nicht mit 'U'pdates abgeschaltet.

Durch die Wahl einer verlängerten Pause zwischen den Auslösungen kann die Belichtungszeit unabhängig von der Verzögerung durch die Bildspeicherung exakt definiert werden auch wenn kein Shutter-Signal zur Verfügung steht (s.u.).
Voreingestellt ist eine Pause von 0.2 Sekunden, die aber von der Kamera bei Bedarf verlängert wird - auf Kosten der nächsten Belichtung. Bei zu kurzen Pausen startet u.U. nur jede zweite Auslösung!
Mit langen Pausen können auch Timelapse-Reihen aufgenommen werden. Dazu muss für den Timer die Belichtungszeit auf mindestens 1 Sekunde gestellt werden.
 Wenn die Kamera die Belichtungszeit vorgibt, können Timelapse-Reihen auch einfach durch lange Belichtungszeit erzeugt werden. (Die Kamera löst erst wieder aus, nachdem der Timer die letzte Belichtung beendet hat.)

Wenn der Intervalltimer parallel zu einem Shutter-Emulator betrieben wird, erhält er von diesem den exakten Auslösezeitpunkt und korrigiert damit die gewählte Belichtungszeit.
 Der Effekt ist auch daran zu erkennen, dass zusätzlich zu den 16 Sternen, mit denen ablaufende Bulb-Belichtung symbolisiert werden, ein zusätzlicher Stern angezeigt wird (bei kurzen Zeiten auch mehrere).
 Bei schnellen Auslösungen, die mit einer Punktfolge symbolisiert werden, werden Shutter-Auslösungen durch ein ' gekennzeichnet.
 Bei fehlenden Bulb-Auslösungen wird ein Hinweis angezeigt. Außerdem stoppt der Intervalltimer, wenn nach registrierten Shutter-Auslösungen drei aufeinanderfolgende Auslösungen fehlen und schaltet die Kühlung ab, um ein Beschlagen des unbenutzten, kalten Sensors zu vermeiden.

 ...

Taupunkt-Kontrolle:
Wenn ein Umgebungstemperatur- und Feuchtesensor (z.B. ein AHT20) verfügbar ist, wird der Taupunkt berechnet und angezeigt.
Mit den Befehlen "10K" und "11K" lässt sich die Taupunkt-Kontrolle für die Kühlung aus- und einschalten. Sie wird mit einem '~' Zeichen nach der PWM-Anzeige angezeigt. Vorgabe ist 'ein' bzw. die im EEPROM abgespeicherte Einstellung.
Bei eingeschalteter PID-Regelung passt die Taupunkt-Kontrolle die Zieltemperatur an den Taupunkt an. Ohne PID-Regelung wird die Kühlleistung reduziert, wenn Fenstertemperatur unter den Taupunkt fällt.
Dabei wird die Fenstertemperatur aus der Sensortemperaturmessung abgeschätzt, indem ein Korrekturwert, der mit dem "T" Befehl eingegeben wird, zur Sensortemperaturmessung addiert wird.
Die Taupunkt-Kontrolle kann mit dem Befehl "29K" auf die Fensterheizung erweitert werden. Dann wird die Fensterheizung verstärkt oder reduziert, wenn die Fenstertemperatur unter oder über dem Taupunkt liegt.(Mit "20K ... "25K" wird sie wieder abgeschaltet.)

Die Steuerung nimmt ein Protokoll aller Aufnahmen auf, das die Belichtungszeit, die Sensor- und ggf. die Umgebungstemperatur enthält. Es kann bis etwa 500 Einträge erfassen.
*/
void IntervallTimer::ShowHelpOn_s()	{bluePrintln_P(PSTR(" 123s  : Belichtungszeit in Sekunden"));}
void IntervallTimer::ShowHelpOn_m()	{bluePrintln_P(PSTR(" 123m : Belichtungszeit in Minuten"));}
void IntervallTimer::ShowHelpOn_l()	{bluePrintln_P(PSTR(" 123L : Limit für Seriendauer in Minuten, 0 schaltet Limit ab."));}
void IntervallTimer::ShowHelpOn_p()	{bluePrintln_P(PSTR(" 123P : Pause zwischen Frames in 1/10 Sekunden, 0 für Vorgabe."));}
void IntervallTimer::ShowHelpOn_n()	{bluePrintln_P(PSTR(" 123N : Limit für Seriendauer in Frames, 0 schaltet Limit ab."));}
void IntervallTimer::ShowHelpOn_pcnt()  {bluePrintln_P(PSTR(" 12%   : Belichtungszeit um bis zu +/- Prozent zufällig variieren"));}
void IntervallTimer::ShowHelpOn_stat()  {bluePrintln_P(PSTR(" ?  : Zeige aktuellen Status"));}
//void IntervallTimer::ShowHelpOn_()	{}

// Hilfe zu Heizung 
#if defined(HEATER_POWER) || defined(HEATER_PWM)
void IntervallTimer::ShowHelpOn_h()
{
  bluePrint_P(PSTR(" 2nK : Fenster-Heizung, 20=aus, 21"));
# if defined(HEATER_SOFT_PWM) || defined(HEATER_PWM)
  bluePrint_P(PSTR("..25=Stärke 1 bis 5"));
# else
  bluePrint_P(PSTR("=an"));
# endif
# if defined(DEW_AVAILABLE)
  bluePrintln_P(PSTR(", 29=auto"));
# else
  bluePrintln();
# endif
}
#endif

// Hilfe zu Kühlung + Heizung ohne PID
#if !defined(COOLER_PWM) && defined(COOLER_POWER)
void IntervallTimer::ShowHelpOn_kh()
{
  bluePrintln_P(PSTR(" nK   : Kühlung, 0=aus, 1=an"));
  ShowHelpOn_h();
}
#endif

int IntervallTimer::SendHelp(int what)
{
  switch (what) {
    case 1:
      EnterNewLine();		  // frische Zeile
      bluePrintln_P(PSTR(" Intervalltimer Befehle:"));
      bluePrintln_P(PSTR(" !  : Belichtungssequenz starten/beenden"));
      bluePrintln_P(PSTR("      (zeigt 16 Teilintervalle an)"));
      bluePrintln_P(PSTR(" # : Dauer-Belichtung starten / Alles abbrechen"));
    break;
    case 2:
      bluePrintln_P(PSTR("       (nur im Bulb-Mode, zeigt abgelaufene Sekunden an)"));
      bluePrintln_P(PSTR(" + : Belichtungszeit erhöhen"));
      bluePrintln_P(PSTR("  - : Belichtungszeit verringern"));
      ShowHelpOn_s();
    break;
    case 3:
      ShowHelpOn_m();
      bluePrintln_P(PSTR(" > : Erhöhe Limit für Seriendauer"));
      bluePrintln_P(PSTR(" < : Verringere Limit für Seriendauer"));
      ShowHelpOn_l();
    break;
    case 4:
      ShowHelpOn_n();
      ShowHelpOn_p();
      ShowHelpOn_pcnt();
    break;
    case 5:
      ShowHelpOn_stat();
#if defined(COOLER_PWM) || defined(COOLER_POWER) || defined(HEATER_POWER) || defined(HEATER_PWM)
# if defined(COOLER_PWM)
      bluePrintln_P(PSTR(" K : Kühlung, Hilfe"));
# else // nur COOLER_POWER:
      ShowHelpOn_kh();
# endif
#endif
      bluePrintln_P(PSTR(" E : Einstellungen, Hilfe"));
    break;
    case 6:
#if defined(PROTOCOL)
      bluePrintln_P(PSTR(" A : Aufzeichnung, Hilfe"));
#endif
#if defined(SPIFLASH_CS)
      bluePrintln_P(PSTR(" F : Flash-Update, Hilfe"));
#endif
      bluePrintln_P(PSTR(" R : Reset Hilfe"));
    break;
    case 7:
    break;
    case 8:
      bluePrintln_P(PSTR(" H : Zeige diese Hilfe (H oder h, ...)"));
      bluePrintln_P(PSTR(" U : Keine Updates senden"));
      bluePrintln_P(PSTR(" D : Debug-Nachrichten senden"));
    return 0;	  		  // fertig mit Basis-Hilfe
    case settingsHelpPage:
      bluePrintln_P(PSTR(" Einstellungen:"));
      ShowHelpOn_stat();
      bluePrintln_P(PSTR(" 0E : Zeige aktuelle Einstellungen"));
      bluePrintln_P(PSTR(" 1E : Lade Grundeinstellungen"));
    break;
    case settingsHelpPage+1:
      bluePrintln_P(PSTR(" 2E : Lade gespeicherte Einstellungen"));
      bluePrintln_P(PSTR(" 3E : Speichere aktuelle Einstellungen im EEPROM"));
      bluePrintln_P(PSTR(" 4E : Lösche gespeicherte Einstellungen"));
    return 0;	  		  // fertig mit Einst.-Hilfe

#if defined(PROTOCOL)
    case protocolHelpPage:
      bluePrintln_P(PSTR(" 0A : Aufzeichnung anhalten"));
      bluePrintln_P(PSTR(" 1A : Aufzeichnung fortsetzen"));
      bluePrintln_P(PSTR(" 2A : Aufzeichnung löschen + neu starten"));
      bluePrintln_P(PSTR(" 3A : Aufzeichnung anzeigen (Pause mit A)"));
    return 0;	  		  // fertig mit Einst.-Hilfe
#endif

#if defined(SPIFLASH_CS)
    case uploadHelpPage:
      bluePrintln_P(PSTR(" 0F/4F : Anfang des ex/internen Flash anzeigen"));
      bluePrintln_P(PSTR(" 1F : Upload in externes Flash starten"));
      bluePrintln_P(PSTR(" 2F : Upload als Update installieren"));
      bluePrintln_P(PSTR(" 3F : Upload-Bereich des Flash löschen"));
    return 0;	  		  // fertig mit Flash-Hilfe
#endif
    case resetHelpPage:
      bluePrintln_P(PSTR(" 1R : Soft-Reset (= 2E, lade Einstellungen)"));
      bluePrintln_P(PSTR(" 2R : Hard-Reset (nur mit Urboot/Optiboot!)"));
    return 0;	  		  // fertig mit Flash-Hilfe

    case calibHelpPage:
#if defined(COOLER_PWM) || defined(COOLER_POWER)
      bluePrintln_P(PSTR(" Kühlung:"));
# if defined(COOLER_PWM)
#  if defined(COOLER_POWER)
      bluePrintln_P(PSTR(" nK : 0=aus, 1=minimal, 2=niedrig, 3=std, 4=hoch, 5=maximal,"));
#  else
      bluePrintln_P(PSTR(" nK : 0,1=minimal, 2=niedrig, 3=std, 4=hoch, 5=maximal,"));
#  endif
      bluePrintln_P(PSTR("         7: Ziel-Temperatur eingeben, Regelung starten"));
      bluePrintln_P(PSTR("         8: Regelung starten, Temperatur halten"));
# endif
#endif
    break;
    case calibHelpPage+1:
# if defined(COOLER_PWM)
      bluePrintln_P(PSTR("         9: Regelung in 10 min starten + halten"));
#  if defined(DEW_AVAILABLE)
      bluePrintln_P(PSTR("       1n: Taupunkt-Kontrolle: 0=aus, 1=ein"));
#  endif
# endif
# if defined(COOLER_PWM)
      bluePrintln_P(PSTR(" 1nnnK : PWM-Signal = nnn"));
      ShowHelpOn_h();
# endif
    break;
    case calibHelpPage+2:
      #ifdef COOLER_PWM
      bluePrintln_P(PSTR(" 0C : PID-Koeffizienten anzeigen"));
      bluePrintln_P(PSTR(" 1C : ... auf Basiseinstellung"));
      bluePrintln_P(PSTR(" 2C : ... aus EEPROM laden"));
      #ifdef COOLER_PWM
    break;
    case calibHelpPage+3:
      #endif
      bluePrintln_P(PSTR(" 3C : ... im EEPROM speichern"));
      bluePrintln_P(PSTR(" 4C : ... aus EEPROM löschen"));
      bluePrintln_P(PSTR(" 5C : ... eingeben"));
      #endif
    break;
    case calibHelpPage+4:
      #if defined(CALIB_PID)
      bluePrintln_P(PSTR(" nnC : Kalibriere Kühlung: nn = PWM-Sprung"));
      bluePrintln_P(PSTR("     C : Kalibrierung abbrechen / beenden"));
      #endif
      #if defined(T_SENSOR_AVAILABLE) || defined(DEW_AVAILABLE)
      bluePrintln_P(PSTR(" T : Temperaturkorrektur (T, Tu, Taupunkt) eingeben"));
      #endif
    return 0;		  // fertig mit Kalib.-Hilfe

    case settingsPart:		  // Einstellungen sind zu lang:
    return ShowSettingsPart2();

    default:
    return 0;		  // ?
  }
  return what + 1;		  // weiter, wenn Sendepuffer geleert
}

