/****************************************************************
 * SparkFun_Qwiic_Humidity_AHT20.h
 * SparkFun Qwiic Humidity AHT20 Library Header File
 * Priyanka Makin @ SparkFun Electronics
 * Original Creation Date: March 25, 2020
 * https://github.com/sparkfun/SparkFun_Qwiic_Humidity_AHT20_Arduino_Library
 * 
 * Pickup a board here: https://wwww.sparkfun.com/products/16618
 * 
 * This file prototypes the AHT20 class implemented in SparkFun_Qwiic_Humidity_AHT20.cpp
 * 
 * Development environment specifics:
 *  IDE: Arduino 1.8.9
 *  Hardware Platform: Arduino Uno
 *  Qwiic Humidity AHT20 Version: 1.0.0
 * 
 * This code is lemonade-ware; if you see me (or any other SparkFun employee)
 * at the local, and you've found our code helpful, please buy us a round!
 * 
 ******************************************************************/
/*
* 
* Modified due to special requirements for camera temperature control.
* Erwin Lotter, Nov 15, 2024
* 
*/

#ifndef AHT20C_H
#define AHT20C_H

#include <Arduino.h>
#include <Wire.h>

#define AHT20C_DEFAULT_ADDRESS 0x38
#define AHT20C_SAMPLEPERIOD    2000	  // Millisekunden zwischen regelmäßigen Messungen

enum registers
{
    sfe_aht20_reg_reset = 0xBA,
    sfe_aht20_reg_initialize = 0xBE,
    sfe_aht20_reg_measure = 0xAC,
};

class AHT20C
{
  private:
    uint8_t _deviceAddress;
    bool measurementStarted;          // waiting for new data
    bool confirmed;                   // set once after begin()
    uint16_t nexttriggertime;
    float   humidity;
    int16_t Tx10;

  public:
    explicit AHT20C(const uint8_t deviceAddress = AHT20C_DEFAULT_ADDRESS);

    bool begin();                  //Sets the address of the device
    bool isConfirmed()  const {return confirmed;}            //Checks success of begin()
    bool needTrigger()  const {return (int16_t)(nexttriggertime - (uint16_t)millis()) < 0;} //Checks if cycle time is expired
    bool triggerMeasurement(); //Trigger the AHT20 to take a measurement
    bool isBusy();             //Returns true if the busy bit is set, false otherwise
    int16_t getTemperatureX10(); // returns temperature times 10 in degrees celcius
    float getHumidity();    // returns humidity in % RH

  protected:
    bool isConnected();                   //Checks if the AHT20 is connected to the I2C bus
    bool available();                     //Returns true if new data is available

    //Measurement helper functions
    uint8_t getStatus();       //Returns the status byte
    bool isCalibrated();       //Returns true if the cal bit is set, false otherwise
    bool initialize();         //Initialize for taking measurement
    void readData();           //Read and parse the 6 bytes of data into raw humidity and temp
    bool softReset();          //Restart the sensor system without turning power off and on

};
#endif
