/****************************************************************
 * SparkFun_Qwiic_Humidity_AHT20.cpp
 * SparkFun Qwiic Humidity AHT20 Library Source File
 * Priyanka Makin @ SparkFun Electronics
 * Original Creation Date: March 25, 2020
 * https://github.com/sparkfun/SparkFun_Qwiic_Humidity_AHT20_Arduino_Library
 * 
 * Pick up a board here: https://www.sparkfun.com/products/16618
 * 
 * This file implements all the functions of the AHT20 class.
 * Functions in this library can return the humidity and temperature 
 * measured by the sensor. All data is communicated over I2C bus.
 * 
 * Development environment specifics:
 *  IDE: Arduino 1.8.9
 *  Hardware Platform: Arduino Uno
 *  Qwiic Humidity AHT20 Version: 1.0.0
 * 
 * This code is lemonade-ware; if you see me (or any other SparkFun
 * employee) at the local, and you've found our code helpful, please
 * buy us a round!
 * 
 * Distributed as-is; no warranty is given.
 ***************************************************************/
/*
* 
* Adapted to specific requirements for camera temperature control.
* Erwin Lotter, Nov 15, 2024
* 
* Während Kameraaufnahmen dürfen keine Delays realisiert werden.
* Deshalb liefern Datenabfragen immer bereits gemessene Werte zurück.
* Um zeitnahe Werte für User-Abfragen zu erhalten, muss der Sensor in
* regelmäßigen Abständen aktiviert werden. Die folgenden Funktionen
* können dafür verwendet werden:
* 
* needTrigger() teilt mit, ob die letzte Messung schon mehr als 
*               AHT20C_SAMPLEPERIOD (= 2000) Millisekunden zurückliegt.
* 
* triggerMeasurement() löst eine neue Messung aus und setzt für 
*               needTrigger() die Zeit bis zur nächsten Messung.
* 
* isConfirmed() gibt den von begin() festgestellten Verbindungsstatus zurück.
* 
* getHumidity()        und
* getTemperatureX10()  geben ohne zu warten die zuletzt gemessenen Werte zurück.
* 
* begin()       verwendet Delays! Initialisiert und triggert die erste Messung.
* isBusy()      zeigt an, ob die zuletzt getriggerte Messung noch läuft.
*/

#include "aht20c.h"



AHT20C::AHT20C(const uint8_t deviceAddress)
{
  _deviceAddress = deviceAddress;
}


bool AHT20C::begin()
{
  confirmed = false;
  if (isConnected() == false)
    return false;

  //Wait 40 ms after power-on before reading temp or humidity. Datasheet pg 8
  delay(40);

  //Check if the calibrated bit is set. If not, init the sensor.
  if (isCalibrated() == false) {
    //Send 0xBE0800
    initialize();

    //Immediately trigger a measurement. Send 0xAC3300
    triggerMeasurement();

    delay(75); //Wait for measurement to complete

    uint8_t counter = 0;
    while (isBusy()) {
      delay(1);
      if (counter++ > 100)
        return (false); //Give up after 100ms
    }

    //This calibration sequence is not completely proven. It's not clear how and when the cal bit clears
    //This seems to work but it's not easily testable
    if (isCalibrated() == false) {
      return (false);
    }
  }
  triggerMeasurement();
  return confirmed = true;
}

//Ping the AHT20's I2C address
//If we get a response, we are correctly communicating with the AHT20
bool AHT20C::isConnected()
{
    Wire.beginTransmission(_deviceAddress);
    if (Wire.endTransmission() == 0)
        return true;

    //If IC failed to respond, give it 20ms more for Power On Startup
    //Datasheet pg 7
    delay(20);

    Wire.beginTransmission(_deviceAddress);
    if (Wire.endTransmission() == 0)
        return true;

    return false;
}

/*------------------------ Measurement Helpers ---------------------------*/

uint8_t AHT20C::getStatus()
{
    Wire.requestFrom(_deviceAddress, (uint8_t)1);
    if (Wire.available())
        return (Wire.read());
    return (0);
}

//Returns the state of the cal bit in the status byte
bool AHT20C::isCalibrated()
{
    return (getStatus() & (1 << 3));
}

//Returns the state of the busy bit in the status byte
bool AHT20C::isBusy()
{
    return (getStatus() & (1 << 7));
}

bool AHT20C::initialize()
{
    Wire.beginTransmission(_deviceAddress);
    Wire.write(sfe_aht20_reg_initialize);
    Wire.write((uint8_t)0x08);
    Wire.write((uint8_t)0x00);
    if (Wire.endTransmission() == 0)
        return true;
    return false;
}

bool AHT20C::triggerMeasurement()
{
    Wire.beginTransmission(_deviceAddress);
    Wire.write(sfe_aht20_reg_measure);
    Wire.write((uint8_t)0x33);
    Wire.write((uint8_t)0x00);
    if (Wire.endTransmission() == 0) {
      nexttriggertime = (uint16_t)millis() + AHT20C_SAMPLEPERIOD;
      measurementStarted = true;
      return true;
    }
    return false;
}

// Loads new measurement if available and not yet read.
void AHT20C::readData()
{
  if (measurementStarted && !isBusy()
      && Wire.requestFrom(_deviceAddress, (uint8_t)6) > 0
  ) {
    Wire.read(); // Read and discard state
    //
    // Die Daten werden in 5 Bytes gelesen, von denen die ersten 20 Bit die
    // Feuchte-, die nächsten 20 Bit die Temperatur-Daten liefern.
    // Die Feuchte-Daten geteilt durch 2^20 ergeben die Feuchte (0..1).
    // Die Temperatur-Daten mal 200 geteilt durch 2^20 ergeben 
    //  (Temperatur + 50) in °C.      (datasheet pg 8)
    //
    //*
    uint16_t w = ((uint16_t)Wire.read() << 8) | (uint16_t)Wire.read();
    uint8_t midByte = Wire.read();
    humidity = float(w) * (100.0 / (1048576>>4));    //unter Verzicht auf die 4 Bits von midByte!
    //oder als int:
    //uint16_t Hx100 = (uint32_t(w) * 25) >> 14;    //unter Verzicht auf die 4 Bits von midByte!
    //uint16_t Hx100 = ((w >> 5) * 25) >> 9;        //(sollte auch keinen Rundungsfehler geben)
    /*/
    uint32_t incoming = (uint32_t)Wire.read() << (8 * 2);
    incoming |= (uint32_t)Wire.read() << (8 * 1);
    uint8_t midByte = Wire.read();
    incoming |= midByte;    //From datasheet pg 8:
    humidity = ((float)(incoming >> 4) / 1048576) * 100;    // 1048576 = 2^20
    //*/

    //*
    uint16_t t;
    t = ((uint16_t(midByte) << 8) | Wire.read()) << 4;    // 4 Bits oben abschneiden
    uint32_t temperature = (uint32_t(t) << 4) | uint8_t(Wire.read()); // (32 Bit gegen Überlauf oder Rundungsfehler notwendig!)
    Tx10 = int16_t((temperature * 125) >> 16) - 500;      // (statt ((T * 12.5) >> 16) - 50)
    /*/
    uint32_t temperature; // 16 Bit reichen nicht wg. Rundungsfehler!
    temperature = (uint32_t)(midByte & 0x0F) << (8 * 2);        //Need to get rid of data in bits > 20
    temperature |= (uint32_t)Wire.read() << (8 * 1);
    temperature |= (uint32_t)Wire.read() << (8 * 0);
    //float tempCelsius = ((float)temperature / 1048576) * 200 - 50;  //From datasheet pg 8
    //Tx10 = int16_t((float)temperature * (2000. / 1048576)) - 500;
    Tx10 = int16_t((temperature * 125) >> 16) - 500;
    //*/
  }
  measurementStarted = false;
}


int16_t AHT20C::getTemperatureX10()
{
  readData();
  return Tx10;
}

float AHT20C::getHumidity()
{
  readData();
  return humidity;
}


bool AHT20C::softReset()
{
    Wire.beginTransmission(_deviceAddress);
    Wire.write(sfe_aht20_reg_reset);
    if (Wire.endTransmission() == 0)
        return true;
    return false;
}

