//////////////////////////////////////////////////////////////////////////////
//		 Arduino Intervall-Timer mit Bluetooth-Modul
//			PID Temperaturregelung
//
//       17.08.2024	                             Erwin Lotter
//////////////////////////////////////////////////////////////////////////////

#include "IntervallTimer.h"


#ifdef T_SENSOR_AVAILABLE
int IntervallTimer::ReadTemperatureX10()
{
  return (int)ReadTemperature_times_10() + tempCorr;
}
#endif


#if defined(AHT20_AVAILABLE)
#include "aht20c.h"

AHT20C aht20;		// braucht 3306 Bytes
#endif

void setupDew() 
{
#if defined(AHT20_AVAILABLE)
  Wire.begin(); //Join I2C bus
  aht20.begin();
#endif
}

bool IntervallTimer::ReadAmbientTemperatureX10(int& temp)
{
#if defined(AHT20_AVAILABLE)
  if (aht20.isConfirmed()) {
    temp = aht20.getTemperatureX10() + TuCorr;
    return true;
  }
#endif
  temp = temp;
  return false;
}


bool IntervallTimer::ReadAmbientHumidity(float& humidity)
{
#if defined(AHT20_AVAILABLE)
  if (aht20.isConfirmed()) {
    humidity = aht20.getHumidity();
    return true;
  }
#endif
  humidity = humidity;
  return false;
}

// Sensor-Messung anstoßen
void IntervallTimer::TriggerAmbientSensor()
{
#if defined(AHT20_AVAILABLE)
  if (aht20.isConfirmed()) {
    aht20.triggerMeasurement();
  }
#endif
}


// Sensor-Messung neu starten, wenn Zykluszeit abgelaufen
void IntervallTimer::CycleAmbientSensor()
{
#if defined(AHT20_AVAILABLE)
  if (aht20.isConfirmed()  &&  aht20.needTrigger()) {
    aht20.triggerMeasurement();
  }
#endif
}


//
// Calculate dew point temperature times 10 from the given 
//  relative humidity (in %) and the ambient temperature Tu.
//
int IntervallTimer::CalcDewPointX10(const float& humidity, int TuX10)
{
  // https://www.wetterochs.de/wetter/feuchte.html#f1
  // https://loxwiki.atlassian.net/wiki/spaces/LOX/pages/1518403585/Taupunkt+berechnen -->
  // T >= 0°C: Td = 243.12 * ((17.62 * T) / (243.12+T)+LN(humidity/100)) / ((17.62 * 243.12) / (243.12+T)-LN(humidity/100))
  // T < 0°C:  Td = 272.62 * ((22.46 * T) / (272.62+T)+LN(humidity/100)) / ((22.46 * 272.62) / (272.62+T)-LN(humidity/100))
  // 
  // Simplified calculation (https://www.schweizer-fn.de/lueftung/feuchte/feuchte.php):
  //  Td = pow(humidity/100, 1/8.02) * (109.8 + T) -109.8	    // pow() kostet 594 Bytes
  //    ~= sqrt(sqrt(sqrt(humidity/100))) * (109.8 + TuX10/10) -109.8  // minimal ungenauer: sqrt(sqrt(sqrt())) kostet 224 Bytes
  // 
  // Weitere Alternative: https://www.mikrocontroller.net/topic/306226#3292459
  // t in °C,  humidity in %,  x = 1.0 - 0.01 * humidity;
  // Td = T - ((14.55 + 0.114 * T) * x + ((2.5 + 0.007 *T) * x)^3 + (15.9 + 0.117 * T) * x^14)
  // Td*10 = TuX10 - 10*((14.55 + 0.0114 * TuX10) * x + ((2.5 + 0.0007 *TuX10) * x)^3 + (15.9 + 0.0117 * TuX10) * x^14)
  //
  // Alle Berechnungen liefern sehr ähnliche Ergebnisse außer für sehr geringe
  // Feuchten (< 3%), die nicht relvant sind.
  // Größere Abweichungen gibt es für Eiskondensation, die früher einsetzt, aber 
  // auch hier nur für geringe Feuchte (<40%) und starke Unterkühlung (dT < -10°C).
  //

  return int(sqrt(sqrt(sqrt(humidity/100))) * (1098 + TuX10)) -1098;   // gleich x10 berechnen
  
  // kostet 622 Bytes:
  //return int(pow(humidity/100, 0.124688)) * (1098 + TuX10)) -1098;   // gleich x10 berechnen
  
  /* kostet 290 Bytes:
  float x = 1.0 - 0.01 * humidity;
  float t = ((2.5 + 0.0007 *TuX10) * x);
  t = t*t*t + (14.55 + 0.0114 * TuX10) * x;
  // ^14 == ^8 * ^6:
  x = x*x;    // ^2
  float x4 = x*x; // ^4
  t += (15.9 + 0.0117 * TuX10) * x * x4 * x4 * x4;	// ^2 * ^4 * ^4 * ^4
  return TuX10 - int(10 * t);
  //*/
}


////////////////////////////////////////////////////////////////////////////////////////
// Set cooling mode 1 (lowest) to 5 (highest)
//  according to commands "1K" .. "5K". Initial mode (= 3) is applied for mode -1.
// 
// Returns the calculated COOLER_PWM output in the range 0 - 255
//  or 1 if PWM is not available.
//
inline int ApplyCoolingMode(int mode)
{
#ifdef COOLER_PWM
  if (mode == -1) {
    mode = 3;
  }
  int pwmout = CoolerPWMLevels[
    #if COOLER_PWM_REVERSE    // The cooling versus PWM signal curve has a negative slope:
      4 -		//  0..4 --> 4..0
    #endif
      (mode-1)];
  return pwmout;
#else
  mode = mode;
  return 1;
#endif 
}



#if defined(COOLER_PWM) || defined(COOLER_POWER)
//
// Set cooling mode by calling ApplyCoolingMode() and write the result
//  to coolingLevel and the appropriate output(s).
// 
//  -1:  Default cooling
//  0:   Switch cooling off
//  1-5: Set cooling level according to commands "20K" .. "25K". 
//  9:   Cooling controlled by dew protection logic
// 
// 'coolinglevel' receives the output value, 0 switches cooling off.
// 'coolingMode':
//  0: Cooling is always on (if coolinglevel != 0)
//  tmfAuto (=0x01): Cooling is controlled by dew protection logic
//  tmfOff (=0x80): Cooling is off 
// (tmfOff and tmfAuto can't be combined!)
// 
void IntervallTimer::SelectCoolingMode(int selection)
{
  if (selection == 0) {
    coolingMode = tmfOff;
    SetCoolingLevel(0);
  }
  else {
    StartCooling(ApplyCoolingMode(selection));
  }
  #if defined(DEW_AVAILABLE)
  dewLimit = coolingLevel;	// Grenze für Dew-Kontrolle
  #endif
}


//
// Set the cooling level and write to COOLER_PWM and/or COOLER_POWER..
//
void IntervallTimer::SetCoolingLevel(uint8_t level)	    
{
  coolingLevel = level; 
  WriteCoolingLevel();
}


void IntervallTimer::StartCooling(uint8_t level)	    
{
  coolingMode &= ~tmfOff;  // Cooling is on
  SetCoolingLevel(level);
  #if defined(DEW_AVAILABLE)
  dewLimit = level;	// Grenze für Dew-Kontrolle
  #endif
}


/* Klappt ohne Extra-Code auch mit int-Parameter:
inline int ApplyCoolerPWMDirection(int8_t delta)
{
  return
    #if COOLER_PWM_REVERSE  // the cooling versus PWM output signal has a
    -			  //  negative slope (i.e. higher output = less cooling)?
    #endif
    delta;
}
*/
inline int ApplyCoolerPWMDirection(int delta)
{
  return
    #if COOLER_PWM_REVERSE  // the cooling versus PWM output signal has a
    -			  //  negative slope (i.e. higher output = less cooling)?
    #endif
    delta;
}


//
// Increase the cooling level by step (or decrease if step < 0).
//
void IntervallTimer::IncreaseCoolingLevelBy(int8_t step)
{
  SetCoolingLevel((uint8_t)max(0, min(255, (int)coolingLevel + ApplyCoolerPWMDirection(step))));
}

#endif

void IntervallTimer::WriteCoolingLevel()
{
  #ifdef COOLER_PWM
  analogWrite(COOLER_PWM, coolingLevel);
  #endif
  #ifdef COOLER_POWER
  digitalWrite(COOLER_POWER, coolingMode<0 ? !COOLER_POWER_ON : COOLER_POWER_ON);
  #endif
}

//
// Translate window heating mode:
//  1-5: Set level (1=lowest, 5=highest) according to commands "20K" .. "25K". 
//  -1:  Default heating
// 
// Returns
//  the calculated PWM value in the range 0 - 255
//  or 0/1 if no PWM is avl. (w/o HEATER_SOFT_PWM)
//
uint8_t ApplyHeatingMode(int mode)
{
#if defined(HEATER_PWM) || defined(HEATER_SOFT_PWM)
  if ((unsigned)mode > 5) {   //  -1 --> default value
    mode = 5;
  }
  // PWM settings for "21K" to "25K" (in reverse order if reverse PWM signal)
  return HeaterPWMLevels[(mode-1)];
#else
  mode = mode;
  return 1;
#endif 
}


#if defined(HEATER_PWM) || defined(HEATER_POWER)
//
// Set heating mode by calling ApplyHeatingMode() and write result
//  to heatingLevel and the appropriate output(s).
// 
//  -1:  Default heating
//  0:   Switch heating off
//  1-5: Set heating level according to commands "20K" .. "25K". 
//  9:   Heating controlled by dew protection logic
// 
// 'heatinglevel' receives the output value, 0 switches heating off.
// 'heatingMode':
//  0: Heating is always on (if heatinglevel != 0)
//  tmfAuto (=0x01): Heating is controlled by dew protection logic
//  tmfOff (=0x80): Heating is off 
// (tmfOff and tmfAuto can't be combined!)
// 
void IntervallTimer::SelectHeatingMode(int selection)
{
  uint8_t level;
  if (selection == 0) {
    heatingMode = tmfOff;
    level = 0;
  }
  else {
    heatingMode = 0;
    if (selection == 9) {	  // control by dew protection logic
      heatingMode = tmfAuto;      //  (schaltet ggf. auch ein!)
      selection = -1;
    }
    level = ApplyHeatingMode(selection);
  }
  SetHeatingLevel(level);
}


//
// Change the heating level Lo-Byte and output it to HEATER_PWM (Soft-PWM will be actuated automatically).
// The HiByte remains unchanged and indicates special cases:
//  0: Heating is always on (if LoByte != 0)
//  tmfAuto (=0x01): Heating is controlled by dew protection logic
//  tmfOff (=0x80): Heating is off 
// (tmfOff and tmfAuto can't be combined!)
//
inline
void IntervallTimer::SetHeatingLevel(uint8_t level)
{
  heatingLevel = level; 
  WriteHeatingLevel();
}


//
// Increase the cooling level by step (or decrease if step < 0).
//
void IntervallTimer::IncreaseHeatingLevelBy(int8_t step)
{
  SetHeatingLevel((uint8_t)max(0, min(255, (int)heatingLevel + (int)step)));
}
#endif

//
// Write the given signal to the Heating PWM output except for 
//  Soft-PWM, which will be actuated automatically!
// 
void IntervallTimer::WriteHeatingLevel()	    
{
#if !defined(HEATER_SOFT_PWM)
  #if defined(HEATER_PWM)
    #if defined(HEATER_PWM_REVERSE) && HEATER_PWM_REVERSE    // The heating versus PWM signal curve has a negative slope:
    analogWrite(HEATER_PWM, 255-heatingLevel);
    #else
    analogWrite(HEATER_PWM, heatingLevel);
    #endif
  #endif
  #if defined(HEATER_POWER)
  digitalWrite(HEATER_POWER, IsHeating()  &&  heatingLevel != 0 ? !HEATER_POWER_ON : HEATER_POWER_ON);
  #endif
#endif
}

//////////////////////////////////////////////////////////////////////////////////////////

//dew point = and relative humidity of the air. Multiply the temperature by 17.625 and divide the result by the temperature plus 243.04.
//////////////////////////////////////////////////////////////////////////////////////////
#define PID_Cycle 1000	  // delay between PID actuations in milliseconds - don't change!

//
// Update cooling PID and heating soft-pwm
// 
// Return action on cooling PID:
//  1: just updated 
//  0: not yet updated
// -1: not active
//
int IntervallTimer::UpdateCooling()
{
#if defined(DEW_AVAILABLE)
  CycleAmbientSensor();		// neue Messung auslösen, wenn Zykluszeit (2s) um ist
#endif

#if defined(COOLER_PWM) || defined(DEW_AVAILABLE) || defined(HEATER_SOFT_PWM)
  uint32_t time = millis();	// millis() braucht 29 Takte (1.8µs @ 16MHz), eine long-Zuweisung aus/in Register 4x2 Takte
#endif

#if defined(DEW_AVAILABLE)
  uint8_t  secnds = uint8_t(time >> 10);   // Zeit in Sekunden (Überlauf nach 256s!)
  //
  // Taupunktkontrolle via Kühlung oder Heizung:
  // Liest alle 10s Temperatur und Feuchte und berechnet den Taupunkt.
  // Verringert oder erhöht die Heizleistung, wenn Sensor wärmer oder
  //  kälter als Taupunkt.
  // Korrigiert die PID Zieltemperatur, wenn unter Taupunkt.
  //  Ohne PID wird die Kühlleistung direkt angepasst.
  //
  bool dewH = IsDewHeatingActive()
    #if HEATER_POWER_BY_COOLER	  // no heating if cooler and fan power are off:
     && IsCooling()
    #endif
    ;
  bool dewC = IsDewControlActive();
  if (dewH || dewC) {  // aktiv?
    if (secnds >= nextDewCheck) {			// nächste Kontrolle?
      int Tu;
      if (ReadAmbientTemperatureX10(Tu)) {
	float humidity;
	if (ReadAmbientHumidity(humidity)) {	// neuen Taupunkt eintragen, wenn größer als Zieltemp.
	  //humidity = 90;  // Testwert für hohe RF setzen
	  int dewptx10 = CalcDewPointX10(humidity, Tu) - TdCorr;		// auf Sensor korrigierter Taupunkt
	  int8_t delta = int8_t(ReadTemperatureX10() - dewptx10);	// Unterschied in 1/10° 
	  delta = max(-10, min(delta, 10));				// T-Unterschied begrenzen auf +/- 10/10°
	  //
	  // heatingLevel an den aktuellen Abstand zwischen Sensor- und
	  //  Taupunkttemperatur anpassen:
	  // Alle 10 Sekunden wird diese Routine durchlaufen. Damit eine
	  //  komplette Umschaltung in 10 Minuten ablaufen kann, müssen wir 
	  //  je Durchlauf 1/60-tel des Regelbereichs, also 256/60 ~= 4, ändern.
	  //
	  int8_t step = delta >> (3 - 2);   // damit den Regelschritt (= 4) nach unten skalieren, wenn |dT| < 1° (bzw. 1.25°)
	  
	  if (dewH) {	    // Heizung an Taupunkt anpassen?
	    //
	    // Die Korrektur ist > 0, wenn der Sensor wärmer ist. Da dann
	    //  die Heizung reduziert werden muss, wird sie abgezogen:
	    //
	    #if defined(HEATER_PWM) || defined(HEATER_SOFT_PWM)
	    IncreaseHeatingLevelBy(-step) ;
	    #else
	    SetHeatingLevel(step > 0 ? 0 : 1) ;
	    #endif
	    if (debug) {
	      PrintHeaterStatus();
    	    }
	  }

	  if (dewC) {		    // Taupunktkontrolle via Kühlung?
	    #if defined(COOLER_PWM)
	    if (IsPIDActive()) {	    //  mit lfd. PID:
	      // Erhöht die PID Zieltemperatur, wenn sie den Taupunkt unterschreitet.
	      //
	      PID_setTemp = max(dewptx10, PID_svdTemp);
	    }
	    else 
	    #endif
	    { // Kühlung verringern, wenn der Sensor kälter ist (step < 0)
	      //  und erhöhen, wenn umgekehrt
  	      IncreaseCoolingLevelBy(step);
	      if (coolingLevel
		    #if COOLER_PWM_REVERSE
		    <
		    #else
		    >
		    #endif
		      dewLimit
	      ) {
		coolingLevel = dewLimit;  // begrenzen auf die voreingestellte Kühlung
	      }
	      if (debug) {
		PrintCoolerStatus();
    	      }
	    }
	  }
	}
      }
      nextDewCheck = secnds + 10;	    // alle 10s
    }
  }
#endif

#if defined(HEATER_SOFT_PWM)
  //
  // Divide time in ~ one-second slots. 
  // Divide each slot in 256 slices which get high or low depending on 
  //  heatingLevel setting.
  // millis() >> 10 --> 1 slot
  // millis() >> (10-8) --> 1 slice
  // (millis() >> (10-8)) & 0x00FF --> slice mask
  //
  bool on = uint8_t((time >> (10 - 8))) < heatingLevel; // switch ON as long as slice has not reached level
  digitalWrite(HEATER_POWER, on ? HEATER_POWER_ON : !HEATER_POWER_ON);
#endif

#if !defined(COOLER_PWM)
  return -1;
#else
  #if defined(CALIB_PID)
  if (IsCalibPID()) {    // PID calibration running?
    return UpdatePIDCalib(false);
  }
  #endif
  if (PID_delay < 0) {
    return -1;		    // not startet
  }
  if (PID_delay > 0) {	    // still waitung for start?
    if (timeReached(PID_prevTime)) {
      InitPID(ReadTemperatureX10());
    }
    else {
      return -1;	    // still waiting
    }
  }
  //
  // PID Controller is running:
  //
  int32_t elapsedTime = time - PID_prevTime;
  if (elapsedTime < PID_Cycle) {	  // wait a second until next cycle
    return 0;
  }
  PID_prevTime = time;
  int Tsens = ReadTemperatureX10();

  int error = ApplyCoolerPWMDirection(PID_setTemp - Tsens);

  #define kp  KpKiKd[0]
  #define ki  KpKiKd[1]
  #define kd  KpKiKd[2]
  float PID_p = kp * error;

  PID_i = PID_i + (ki * error); // *(elapsedTime/PID_Cycle)    // assume 1 

  float PID_d = kd * (error - PID_error); // /(elapsedTime/PID_Cycle)    // assume 1 

  int newCoolingLevel = (int)(PID_p + PID_i + PID_d);
  if (debug) {
    bluePrintTdegC(Tsens);
    bluePrint_P(PSTR(", p="));
    bluePrint((int)PID_p);
    bluePrint_P(PSTR(" + i="));
    bluePrint((int)PID_i);
    bluePrint_P(PSTR(" + d="));
    bluePrint((int)PID_d);
    bluePrint_P(PSTR(" -> "));
    bluePrintln(newCoolingLevel);
  }
# if defined(COOLER_POWER)	// if there is a data line to deaktivate peltier we can use it in
  bool coff = false;		//  case the PWM output is on minimum and the temp is still too low
# endif
  if (newCoolingLevel < 0) {    // limit to PWM range 0 - 255:
    #if defined(COOLER_POWER) && !COOLER_PWM_REVERSE
    //
    // the cooling versus PWM output signal has a positive slope:
    // 
    if (newCoolingLevel < -15) {
      coff = true;		// switch cooling off
      PID_i = PID_i - (ki * error); // preserve last on-state
    }
    #endif
    newCoolingLevel = 0;	// limit PWM output
  }
  if (newCoolingLevel > 255) {
    #if defined(COOLER_POWER) && COOLER_PWM_REVERSE  // the cooling versus PWM output signal has a negative slope:
    if (newCoolingLevel > 270) {
      coff = true;		// switch cooling off
      PID_i = PID_i - (ki * error); // preserve last on-state
    }
    #endif
    newCoolingLevel = 255;	// limit PWM output  
  }
  SetCoolingLevel((uint8_t)newCoolingLevel);
# if defined(COOLER_POWER)
  if (coff != PID_coolingOff) {	    // switch cooling off if lowest level ist still too high
    bool wasDewC = IsDewControl();
    SelectCoolingMode(coff ? 0 : 1);   //  or back to lowest level
    PID_coolingOff = coff;
    if (wasDewC) {
      coolingMode |= tmfAuto;
    }
  }
# endif
  PID_error = error;     // previous error --> next loop.
  return 1;
#endif
}


#ifdef COOLER_PWM
//////////////////////////////////////////////////////////////////////////
// Start PID temperature control after given delay
//
void IntervallTimer::StartPID(int afterSeconds)
{
  StopPID(); 
  if (!IsCooling()) {		  // activate std cooling, if cooling is off
    SelectCoolingMode(-1);
  }
  PID_delay = afterSeconds+1;		  // status: PID
  PID_prevTime = millis() + afterSeconds * 1000L; // set delay
}


//
// Start PID temperature control for given destination tempertureX10.
//
void IntervallTimer::StartPIDT(float tempX10)
{
  StopPID(); 
  if (!IsCooling()) {		  // activate std cooling, if cooling is off
    SelectCoolingMode(-1);
  }
  InitPID((int)(tempX10 + tempCorr + 0.5));
}


//
// Abort any temperature PID action w/o notification
//
void IntervallTimer::AbortPID()
{
  PID_delay = -1;
}


//
// Stop temperature PID
//
void IntervallTimer::StopPID()
{
  #if defined(CALIB_PID)
  if (IsCalibPID()) {
    //UpdatePIDCalib(true);
  }
  else 
  #endif
  {
    if (PID_delay >= 0) {
      EnterNewLineLn(PSTR("Temperaturregelung abgebrochen"));
      if (PID_coolingOff) {
        SelectCoolingMode(1);		  // re-activate (minimal) cooling if switched off by PID control
      }
    }
    AbortPID();
  }
}

// PID controller in action atm?
bool IntervallTimer::IsPIDActive() const 
{
  return PID_delay == 0; 
}

// PID inactive and not waiting and not calibrating?
bool IntervallTimer::IsPIDOff() const 
{
  return PID_delay == -1; 
}


void IntervallTimer::DeletePID() 
{
  #if defined(CALIB_PID)
  if (IsCalibPID()) {
    AbortPID();
  }
  delete PIDCal; 
  PIDCal = 0;
  #endif
}


//
// Taupunktkontrolle umschalten
//
  #if defined(DEW_AVAILABLE)
void IntervallTimer::EnableDewControl(bool en)	
{
  if (IsDewControl() != en) {
    if (en) {	  		  // ein:
      PID_svdTemp = PID_setTemp;  //  Zieltemperatur sichern
      nextDewCheck = uint8_t(millis() >> 10);	// Start jetzt (Zeit in Sekunden)
      coolingMode |= tmfAuto;
    }
    else {			  // aus:
      PID_setTemp = PID_svdTemp;  //  Zieltemperatur wiederherstellen
      coolingMode &= ~tmfAuto;
    }
  }
}
  #endif

//
// Init PID settings
// Achtung: Zu tempX10 ist schon der Korrekturwert der T Messung (tempCorr)
//  addiert. Dieser muss auch bei User-Angabe der Zieltemperatur addiert
//  werden!
// Bei der Anzeige von PID_setTemp muss tempCorr abgezogen werden!
//
void IntervallTimer::InitPID(int tempX10)
{
  PID_coolingOff = false;
  PID_delay = 0;	    // srart now
  PID_setTemp = PID_svdTemp = tempX10;	// Zieltemp. eintragen und Sicherung davon nach PID_svdTemp
  #if defined(DEW_AVAILABLE)
  nextDewCheck = 1;
  #endif
  PID_i = coolingLevel;	    // equilibrium cooling level
  PID_error = 0;
  PID_prevTime = millis() - PID_Cycle; // sofort anfangen
  PrintPIDTemp();
  bluePrintln();
}


#define TEST_PIDCALIB 0
#if TEST_PIDCALIB
  // Testdaten für die PID-Kalibrierung
  // T in 0.1°C im Abstand von 60s
  const int testT[90] = {
    287, 285, 281, 275, 272, 268, 265, 262, 260, 257, 254, 252, 249, 247, 245, 
    243, 241, 239, 238, 236, 234, 233, 232, 230, 229, 228, 226, 225, 224, 
    223, 222, 221, 220, 220, 219, 218, 217, 216, 216, 215, 215, 214, 213, 
    213, 212, 212, 212, 211, 211, 210, 210, 210, 209, 209, 209, 208, 208, 
    207, 206, 205, 204, 204, 204, 203, 203, 203, 203, 203, 203, 203, 203, 
    203, 203, 203, 202, 202, 202, 202, 202, 202, 202, 202, 202, 202, 202, 
    202, 202, 202, 202, 202, 
  };
  int testI;	    // index in testT
  uint32_t nexttestmilli;
# define test_dt 600	    // msec between test samples (real value = 60000)
#endif

#if defined(CALIB_PID)
//
// Start PID calibration with given output step
//
bool IntervallTimer::CalibPID(int step)
{
  #if !TEST_PIDCALIB
  if (step < 10) {
    bluePrintln_P(PSTR("Kühlungs-Delta zu klein für PID-Kalibrierung!"));
    return false; 
  }
  if (!IsCooling()) {
    SelectCoolingMode(-1);
  }
  int newCoolingLevel;
  bool badstep;
  #if COOLER_PWM_REVERSE
    newCoolingLevel = (uint16_t)coolingLevel + step;  // the cooling versus PWM output signal has a negative slope 
    badstep = newCoolingLevel < 0;	    //  (i.e. higher output = less cooling): increase cooling by negative step
  #else
    newCoolingLevel = (uint16_t)coolingLevel + step;
    badstep = newCoolingLevel > 255;
  #endif
  if (badstep) {
    bluePrintln_P(PSTR("Kühlungs-Delta zu groß für PID-Kalibrierung!"));
    return false; 
  }
  #endif
  if (!PIDCal) {
    PIDCal = new StepResponse<100, RSCAL>;
    if (!PIDCal) {
      bluePrintln_P(PSTR("Nicht genug Speicher für die PID-Kalibrierung!"));
      return false; 
    }
  }
#if TEST_PIDCALIB
  testI = 0;
  nexttestmilli = millis() + test_dt;	  // supply reading every test_dt msec
  PIDCal->Init(1, -100);
  step = step;
  bluePrint_P(PSTR("PID-Kalibrierungstest gestartet"));
#else
  // The sign of the measure T will be inverted to get a increasing response.
  // To compensate this, step needs to be negated, too.
  // (Having applied a negative step to the coolingLevel in case 
  //  of COOLER_PWM_REVERSE != 0 is not taken into account, since this will
  //  be done automatically during temperature control, too.)
  //
  PIDCal->Init(initialResponseDelta_xS, -step);
  bluePrint_P(PSTR("PID-Kalibrierung gestartet, PWM="));
  bluePrint(coolingLevel);
  bluePrint_P(PSTR(" -> "));
  SetCoolingLevel((uint8_t)newCoolingLevel);
  bluePrintln(newCoolingLevel);
#endif
  PID_calCnt = PIDCal->GetCount();
  PID_delay = -2;	  // Flag Calib
  return true;
}


//
// Read again data for PID calibration, finish if stop==true.
//
int IntervallTimer::UpdatePIDCalib(bool stop)
{
  unsigned dt;
  if (!stop) {
  #if TEST_PIDCALIB
    //
    // Test-Updates nur im vorgegebenen Zeitraster ausführen. (Dazwischen
    //  müsste mit einer Auflösung < 0.1°C interpoliert werden.)
    //
    if (nexttestmilli > millis()) {   // wait for next test sample 
      return 2;
    }
    if (testI < (int)ELEMENTS(testT) - 1) {
      testI++;				// next test value
    }
    nexttestmilli += test_dt;
    for (int i = 0; i < 8; i++) {	// PIDCal->Update() builds mean of 8!
      dt = PIDCal->Update(-testT[testI]);  // time units (128 msec) expired since previous update
    }
  #else
    dt = PIDCal->Update(-ReadTemperature_xS());  // time units (128 msec) expired since previous update
  #endif
  }
  if (stop  ||  dt >= (int)(600L * 1000 / 128)) { 	// stop 600s w/ new sample
    //
    // finish if user break or more than 5 minutes since previous update:
    // 
    PID_delay = -1;			// stop calibrating
    calibState = 0;
  #if !TEST_PIDCALIB
    IncreaseCoolingLevelBy(PIDCal->GetOutputStep());	// return to pre-calib cooling level
  #endif
    int r = PIDCal->CalculateCoeffs();
    if (r == 0) {
      bluePrintln_P(PSTR("\nPID-Kalibrierung fehlgeschlagen"));
    }
    else {
      calibState = 4 | r;			    // Rest stückweise ausführen
    }
    return -1;
  }
  if (PIDCal->GetCount() != PID_calCnt) {
    PID_calCnt = PIDCal->GetCount();
    if (debug) {
      blueSerial.print(PID_calCnt);
      bluePrint_P(PSTR(": T="));
      bluePrintTdegC(-PIDCal->GetLastReading_x1());
      bluePrint_P(PSTR(", dt="));
      bluePrintDecimals((millis() - PIDCal->GetStartMillis()) / 100, 1);
      bluePrintln_P(PSTR("s"));
    }
    else {
      bluePrint_P(PSTR("."));	// indicate next response level reached
    }
  }
  return 2;		    // calibrating
}


//
// After PID calibration, let user decide what to do.
//
void IntervallTimer::ContinueCalib()
{
  switch (calibState & ~3) {
    case 4:
      if (calibState & 1) {
	bluePrint_P(PSTR("\nPID-Kalibrierung (PWM="));
	bluePrint(coolingLevel);
	bluePrint_P(PSTR(" -> "));
	bluePrint((int)coolingLevel - ApplyCoolerPWMDirection(PIDCal->GetOutputStep()));
	bluePrint_P(PSTR("):\n1.) "));
	PrintPID(PIDCal->Get_T_sum());
	bluePrint_P(PSTR(" (nach I-TAE) anwenden\n2.) ... dauerhaft eintragen"));
      }
    break;
    case 8:
      if (calibState & 2) {
        bluePrint_P(PSTR("\n3.) "));
	PrintPID(PIDCal->Get_ZieglerNichols());
        bluePrint_P(PSTR(" (Ziegler–Nichols) anwenden\n4.) ... dauerhaft eintragen"));
      }
    break;
    case 12:
        bluePrint_P(PSTR("\n5.) Neu anzeigen\nAuswahl (0 um abzubrechen): "));
    break;
    case 16: {
      if (!blueSerial.available()) {	// Zeichen über BT da?
	return;				//  nein
      }
      uint8_t z = blueSerial.read() - '0';	// nur 0 - 4 akzeptieren
      if (z == 5) {
	PID_calCnt = 0;			// neu anzeigen  
	break;
      }
      if (	z > 4  
	  ||  ((z == 1 || z == 2)  &&  (calibState & 1) == 0)
	  ||  (z >= 3  &&  (calibState & 2) == 0)
	) {
        calibState -= 4;      // query again
	return;
      }
      calibState = 0;
      float* kkk = PIDCal->Get_T_sum();	  // für z=1,2
      switch (z) {
	case 0:	
	  bluePrintln_P(PSTR("\nPID-Kalibrierung abgebrochen\n"));
	  return;
	case 3:
	case 4: kkk = PIDCal->Get_ZieglerNichols(); break;
      }
      for (int i = 0; i < 3; i++) {
	KpKiKd[i] = kkk[i];
      }
      if (z == 2  ||  z == 4) {	  	  // write to EEPROM
	SavePIDtoEEPROM();
        bluePrintln_P(PSTR("\nPID-Koeffizienten im EEPROM gespeichert\n"));
      }
      else {
        bluePrintln_P(PSTR("\nPID-Kalibrierung abgeschlossen\n"));
      }
      return;
    }
    case 20: {			  // neu anzeigen:
      if (PID_calCnt < PIDCal->GetCount()) {
	blueSerial.print(PID_calCnt);
	bluePrint_P(PSTR(": T="));
	bluePrintTdegC(-PIDCal->GetReading_x1(PID_calCnt));
	bluePrint_P(PSTR(", dt="));
	bluePrintDecimals((int)(PIDCal->GetTime(PID_calCnt) * 1.28), 1);// *128 /1000 *10
	bluePrintln_P(PSTR("s"));
	PID_calCnt++;
	return;
      }
      calibState -= 16;	    // Koeffs neu anzeigen
      return;
    }
  }
  calibState += 4;
}
#endif


//
// Load PID coeffs from EEPROM or basic value set.
//
// Returns truee if loaded from EEPROM
//
bool IntervallTimer::ResetPID(bool fromEEPROM)
{
  if (fromEEPROM && LoadPIDfromEEPROM()) {
    return true;
  }
  KpKiKd[0] = PID_Kp;
  KpKiKd[1] = PID_Ki;
  KpKiKd[2] = PID_Kd;
  return false;
}


//
// Print PID coeffs
//
void IntervallTimer::PrintPID(float* kkk)
{
  bluePrint_P(PSTR("Kp = "));
  bluePrintDecimals((int)(kkk[0] * 100), 2);
  bluePrint_P(PSTR(", Ki = "));
  bluePrintDecimals((int)(kkk[1] * 1000), 3);
  bluePrint_P(PSTR(", Kd = "));
  if (fabs(kkk[2]) > 300) {
    bluePrintDecimals((int)(kkk[2] * 10), 1);
  }
  else {
    bluePrintDecimals((int)(kkk[2] * 100), 2);
  }
}

#endif // COOLER_PWM
