/////////////////////////////////////////////////////////////////////////////////////////////
//	ProtocolStack für Arduino Intervall-Timer mit Bluetooth-Modul
// 
//       18.10.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef protocol_h_
#define protocol_h_


// <-- Comment out to force protocol into internal RAM
#if defined(SPIFLASH_CS)	// external Flash available:
#define PROTOCOL_FLASH		// (saves 360 Bytes of program storage space and 506 bytes of dynamic memory!)
#endif
//*/

#if defined(PROTOCOL_FLASH)
#include "FlashUpdate.h"
# define PROTOCOL_STACK_SIZE   0xFF00
#else
# define PROTOCOL_STACK_SIZE   500     // reduce if Arduino gets out of SDRam
  /*
  Jeder Eintrag in ProtocolStack enthält den Frameindex der Aufnahme seit dem Einschalten
  und für jeden Sensor einen Temperturwert in der Reihenfolge T_SENSOR_AVAILABLE, AHT20_AVAILABLE.
  Jeder der Einträge belegt ein int16_t, es sei denn, er wird komprimiert.
  (Protokoll im ext. Flash wird nicht komprimiert.)

  Sobald mindestens ein Wert komprimiert wird, wird ein komprimiertes Bytes (K-Byte)
  gespeichert, das immer daran zu erkennen ist, dass sein MSB gesetzt ist.

  - Wenn der Frameindex gleich dem vorherigen plus eins ist, entfällt sein Eintrag
    und das erste Bytes ist ein K-Byte. 
    Andernfalls ist das erste Wort der Frameindex (max. 0x7FFF). 
    Wenn diesem kein K-Byte folgt, ist es der erste, unkomprimierte Temperturwert.
    Wenn dann nochmals kein K-Byte folgt, ist das der zweite unkomprimierte Temperturwert.
  
  - Wenn sich ein Temperturwert um maximal +/-d von seinem Vorgänger unterscheidet,
    wird die Differenz in n Bits des K-Bytes geschrieben:
     Für nur einen Temperturwert sind das Bit 6-0 und d = 63.
     Für zwei Temperturwerte erhält T1 Bit 6-3 mit d = 7 und T2 erhält Bit 2-0 mit d = 3.
    Der Wert -d-1 zeigt jeweils an, dass ein voller Eintrag notwendig war, der dann
    dem K-Byte folgt. Zur einfacheren und  vorzeichenlosen Verarbeitung wird zur Differenz
    d+1 addiert, so dass dann 0 statt (-d-1) einen Überlauf der Differenz anzeigt.
  */

  #if defined(PROTOCOL_2T)
  const int dT1_shift = 3;
  const int dT1_limit = 8;
  const int dT2_shift = 0;
  const int dT2_limit = 4;
  const int dT2_mask  = (dT2_limit * 2 -1);
  #else
  const int dT1_shift = 0;
  const int dT1_limit = 64;
  #endif
  const int dT1_mask  = (dT1_limit * 2 -1);
#endif

class ProtocolStack {
  public:
	    ProtocolStack() {Reset();}
    void    Reset();
    void    Stop()	      {active = false;}
    void    Continue()	      {active = true;}
    bool    IsActive() const  {return active;}
    int	    GetCount() const  {return cnt;}
    bool    IsFull() const    {return nextWritePos > pbf + maxEntryOfs;}

    bool    IsSending() const {return sending != 0;}
    bool    IsSendingPaused() const	{return sending < 0;}
    bool    StartSending();
    void    ContinueSending()	{sending = 1;}
    void    PauseSending()	{sending = -1;}
    void    StopSending()	{sending = 0;}
    bool    ReadNextLine(int breakCount);

    #if defined(PROTOCOL_2T)
    bool    Add(int frame, int T1, int T2);
    int	    GetRead_T2() const  {return readT2;}
    enum { maxEntrySize = 6 }; // max 6 Bytes
    #else
    bool    Add(int frame, int T1);
    enum    { maxEntrySize = 4 };
    #endif
    int	    GetRead_I() const   {return readIndex;}
    int	    GetRead_T1() const  {return readT1;}

    enum    { maxEntryOfs = PROTOCOL_STACK_SIZE - maxEntrySize };  // noch Platz für einen vollen Eintrag

  protected:
    int     ReadWord();
    void    WriteWord(int w);
    #if !defined(PROTOCOL_FLASH)
    uint8_t ReadByte();
    int	    AppendReadByte(int8_t k_byte);
    void    WriteByte(int8_t b);
    int	    ExtractT1(int8_t k_byte);
    int	    ExtractT2(int8_t k_byte);
    void    DecompressT1(int8_t k_byte);
    void    DecompressT2(int8_t k_byte);
    #endif

  protected:
    #if defined(PROTOCOL_FLASH)
    uint16_t nextWritePos;
    uint16_t nextReadPos;
    uint32_t flashAddress(uint16_t rwa) const	  {return PROTOCOL_FLASHPAGE + uint16_t(rwa);}
    enum   {pbf = 0};
    #else
    int8_t  pbf[PROTOCOL_STACK_SIZE];
    int	    prevIndex;	// previuos index
    int	    prevT1;	// previuos T1
    #if defined(PROTOCOL_2T)
    int	    prevT2;	// previuos T2
    #endif
    int8_t* nextWritePos;
    int8_t* nextReadPos;
    #endif
    int	    cnt;	// index of next entry to add
    int	    readIndex;	// gelesener Index
    int	    readT1;	// gelesene T1
    #if defined(PROTOCOL_2T)
    int	    readT2;	// gelesene T1
    #endif
    bool    active;
    int8_t  sending;
};


//
// Inhalt löschen, Schreibposition auf 0, aktiv setzen.
//
inline
void ProtocolStack::Reset()
{
  sending = 0;
  cnt = 0;
  nextWritePos = pbf;	  // Schreibposition absolut
  #if !defined(PROTOCOL_FLASH)
  prevT1 = 0;		// zuvor geschriebenes T1
  # if defined(PROTOCOL_2T)
  prevT2 = 0;
  # endif
  #endif
  active = true;
}


//
// Ausgabe init.
// Return false wenn nix da.
//
inline
bool ProtocolStack::StartSending()
{
  sending = GetCount() > 0 ? 1 : 0;
  nextReadPos = pbf;	  // Leseposition absolut
  #if !defined(PROTOCOL_FLASH)
  readIndex = 0;	// zuvor gelesener Index 
  readT1 = 0;		// zuvor gelesene T1
  # if defined(PROTOCOL_2T)
  readT2 = 0;		// zuvor gelesene T2
  # endif
  #endif
  return IsSending();
}


#if defined(PROTOCOL_FLASH)
//
// 16-Bit Wert eintragen, Schreibadresse weiter
//
inline
void ProtocolStack::WriteWord(int w)
{
  if ((nextWritePos & 0x0FFF) == 0) {	// neuer 4K-Block?
    bootFlash.eraseSector(flashAddress(nextWritePos));  // sector erase time is 400ms
  }
  bootFlash.writeWord(flashAddress(nextWritePos), w);
  nextWritePos += 2;
}


//
// 16-Bit Wert entnehmen, Leseadresse weiter.
//
inline
int ProtocolStack::ReadWord()
{
  int r = bootFlash.readWord(flashAddress(nextReadPos));
  nextReadPos += 2;
  return r;
}

#else	// ohne Flash, Komprimierung

//
// Byte entnehmen, Leseadresse weiter
//
inline
uint8_t ProtocolStack::ReadByte()
{
  return *nextReadPos++;
}

//
// k_byte und das nächste Byte zu einem signed 15-Bit Wert kombinieren
//  und auf 16 Bit erweitern. 
// Leseadresse weiter.
//
inline
int ProtocolStack::AppendReadByte(int8_t k_byte)
{
  return (int16_t(uint16_t(k_byte) << 9) >> 1) | ReadByte();
}

//
// Signed 15-Bit Wert entnehmen, auf 16 Bit erweitern.
// Leseadresse weiter.
//
inline
int ProtocolStack::ReadWord()
{
  uint8_t r = ReadByte();
  return AppendReadByte(r);
}

//
// Byte eintragen, Schreibadresse weiter
//
inline
void ProtocolStack::WriteByte(int8_t b)
{
  *nextWritePos++ = b;
}

//
// 16-Bit Wert eintragen, Schreibadresse weiter
//
inline
void ProtocolStack::WriteWord(int w)
{
  WriteByte(uint8_t(w >> 8) & 0x7F); // Index komplett schreiben, HiByte zuerst,
  WriteByte(uint8_t(w));	     //  damit das höchste/erste Bit sicher 0 bleibt
}


//
// T1 aus K-Byte entnehmen
//
inline
int ProtocolStack::ExtractT1(int8_t k_byte)
{
  return (k_byte >> dT1_shift) & dT1_mask;
}


#if defined(PROTOCOL_2T)
//
// T2 aus K-Byte entnehmen
//
inline
int ProtocolStack::ExtractT2(int8_t k_byte)
{
  return (k_byte >> dT2_shift) & dT2_mask;
}
#endif

//
// T1 aus K-Byte entnehmen oder unkomprimiert lesen 
//  und in readT1 eintragen.
//
inline
void ProtocolStack::DecompressT1(int8_t k_byte)
{
  int dT1 = ExtractT1(k_byte);
  if (dT1 == 0) {				// T1 ist nicht komprimiert:
    readT1 = ReadWord(); // lesen
  }
  else {
    readT1 += dT1 - dT1_limit;
  }
}

//
// T2 aus K-Byte entnehmen oder unkomprimiert lesen
//  und in readT2 eintragen.
//
inline
void ProtocolStack::DecompressT2(int8_t k_byte)
{
  #if defined(PROTOCOL_2T)
  int dT2 = ExtractT2(k_byte);
  if (dT2 == 0) {				// T2 ist nicht komprimiert:
    readT2 = ReadWord(); // lesen
  }
  else {
    readT2 += dT2 - dT2_limit;
  }
  #else
  k_byte = k_byte;
  #endif
}
#endif

////////////////////////////////////////////////////////////////////////
// Nächste Zeile des Protokolls lesen, Position weitersetzen
//  und 'sending' ggf. abschalten
// Nach jeweils breakCount-1 Aufrufen einmal pausieren.
// 'sending' muss > 0 sein!
// 
// Rückgabe: true, wenn Zeile gelesen
//
inline
bool ProtocolStack::ReadNextLine(int breakCount)
{
  if (nextReadPos < nextWritePos) {   // noch was da?
    if (sending > breakCount) {
      sending = 1;
      return false;
    }
    sending++;
    if (sending < 0) {
      sending = 1;
    }
#if defined(PROTOCOL_FLASH)	// ohne Zwischenpuffer müsste ReadNextLine() nach Ausgabe aufgerufen werden!
    readIndex = ReadWord();
    readT1 = ReadWord();
    #if defined(PROTOCOL_2T)
    readT2 = ReadWord();
    #endif
#else
    int8_t k_byte = ReadByte();
    readIndex++;				// Frameindex 1 weiter
    if (k_byte >= 0) {			        // Frameindex ist nicht komprimiert (ganz am Anfang oder nach Lücke):
      readIndex = (k_byte << 8) | ReadByte();	// lesen
      k_byte = ReadByte();			// nächstes Byte prüfen
      if (k_byte >= 0) {		        // T1 ist nicht komprimiert:
	readT1 = AppendReadByte(k_byte);	//  noch ein Byte dazu
	#if defined(PROTOCOL_2T)
	k_byte = ReadByte();			// nächstes Byte prüfen
	if (k_byte >= 0) {			// T2 ist nicht komprimiert:
	  readT2 = AppendReadByte(k_byte);	//  noch ein Byte dazu
	}
	else {
	  readT2 += ExtractT2(k_byte) - dT2_limit;
	}
	#endif
      }
      else {					// T1 ist komprimiert:
	readT1 += ExtractT1(k_byte) - dT1_limit;
	DecompressT2(k_byte);			// T2 ist evtl. komprimiert
      }
    }
    else {	        		        // Frameindex ist komprimiert:
      DecompressT1(k_byte);			// T1 ist evtl. komprimiert
      DecompressT2(k_byte);			// T2 ist evtl. komprimiert
    }
#endif
  }
  else {
    sending = 0;
  }
  return IsSending();
}


//
// Wert in Protokoll eintragen
//
inline
bool ProtocolStack::Add(int frame, int T1
    #if defined(PROTOCOL_2T)
    , int T2
    #endif
  ) 
{
  if (!IsActive() || IsFull()) {
    return false;
  }
#if defined(PROTOCOL_FLASH)	// ohne Zwischenpuffer müsste ReadNextLine() nach Ausgabe aufgerufen werden!
  WriteWord(frame);
  WriteWord(T1);
  #if defined(PROTOCOL_2T)
    WriteWord(T2);
  #endif
#else
  uint8_t k_byte = 0x80;		// K-Byte erstellen:
  int dT1 = T1 - prevT1;		// T1-Diff + 8 = 1...15, wenn in Bit 6-3 unterzubringen
  if (abs(dT1) < dT1_limit) {		// passt:
    k_byte |= ((dT1 + dT1_limit) << dT1_shift);//  ins K-Byte eintragen
  }
  prevT1 = T1;
  #if defined(PROTOCOL_2T)
  int dT2 = T2 - prevT2;		// T2-Diff + 4 = 1...7, wenn in Bit 2-0 unterzubringen
  if (abs(dT2) < dT2_limit) {		// passt:
    k_byte |= ((dT2 + dT2_limit) << dT2_shift);//  ins K-Byte eintragen
  }
  prevT2 = T2;
  #endif
  prevIndex++;				// Folgeindex
  if (frame != prevIndex) {		// nicht letzter Index + 1:
    prevIndex = frame;
    WriteWord(frame);			//  Index komplett schreiben
    if (abs(dT1) >= dT1_limit) {	// T1 passt nicht:
      WriteWord(T1);			//  komplett schreiben
      #if defined(PROTOCOL_2T)
      if (abs(dT2) >= dT2_limit) {	// T2 passt auch nicht:
	WriteWord(T2);			//  komplett schreiben
      }
      else {				// sonst jetzt das K-Byte
	WriteByte(k_byte);	// K-Byte steht in der Mitte
      }
      #endif
    }
    else {				// sonst jetzt das K-Byte
      WriteByte(k_byte);		// K-Byte steht in der Mitte
      #if defined(PROTOCOL_2T)
      if (abs(dT2) >= dT2_limit) {	// T2 passt auch nicht:
	WriteWord(T2);			//  komplett schreiben
      }
      #endif
    }
  }
  else {				// Index ist 1 weiter:
    WriteByte(k_byte);			// K-Byte steht vorn
    if (abs(dT1) >= dT1_limit) {	// T1 passt nicht:
      WriteWord(T1);			//  komplett schreiben
    }
    #if defined(PROTOCOL_2T)
    if (abs(dT2) >= dT2_limit) {	// T2 passt nicht:
      WriteWord(T2);			//  komplett schreiben
    }
    #endif
  }
#endif
  cnt++;
  return true;
}  
  
#endif
